from __future__ import annotations

from typing import List, Optional, Tuple

from .puzzle import Puzzle
from .puzzle_parsing import PuzzleParser, PuzzlePrinter, PuzzleResizer
from .rle import Rle
from .snapshot import Snapshot


class HexobanSnapshot(Snapshot):
    def __init__(self, moves_data: str = ""):
        from ..game import Tessellation

        super().__init__(tessellation=Tessellation.HEXOBAN, moves_data=moves_data)


class HexobanPuzzle(Puzzle):
    def __init__(self, width: int = 0, height: int = 0, board: Optional[str] = None):
        """
        Arguments:
            width: number of columns
            height: number of rows
            board: If not blank, it will be parsed and board will be created from it,
                ignoring ``width`` and ``height``.
        """
        from ..game import Tessellation

        super().__init__(
            tessellation=Tessellation.HEXOBAN,
            width=width,
            height=height,
            board=board,
            resizer_cls=HexobanPuzzleResizer,
            parser_cls=HexobanPuzzleParser,
            printer_cls=HexobanPuzzlePrinter,
        )


class HexobanTextConverter:
    def convert_to_string(
        self,
        parsed_board: List[str],
        width: int,
        height: int,
        use_visible_floor: bool,
    ) -> str:
        from ..game import index_1d

        floor = Puzzle.VISIBLE_FLOOR if use_visible_floor else Puzzle.FLOOR

        retv: List[str] = []
        for row in range(0, height):
            line = ""
            if row % 2 == 1:
                # beginning half hex for odd rows
                line += floor

            for col in range(0, width):
                puzzle_char = parsed_board[index_1d(col, row, width)]

                line += floor
                if Puzzle.is_empty_floor(puzzle_char):
                    line += floor
                else:
                    line += puzzle_char

            retv.append(line)

        retv = PuzzleParser.normalize_width(retv, floor)
        if self.is_type1(retv):
            retv = self.remove_column_right(retv)

        return "\n".join(retv)

    def is_type1(self, strings: List[str]) -> bool:
        from ..game import Y

        rmnf = self._find_rightmost_non_floor(strings)

        if rmnf is None:
            return False
        elif rmnf > 0:
            y = Y(rmnf, PuzzleParser.calculate_width(strings))
            return y % 2 == 0
        else:  # rmnf == 0
            return True

    @staticmethod
    def _find_rightmost_non_floor(strings: List[str]) -> Optional[int]:
        from ..game import X, Y, index_1d

        def rightmost_finder(strings: List[str], row_parity: int):
            cell_found = False
            height = len(strings)
            width = len(strings[0])
            x = y = 0

            for row in range(row_parity % 2, height, 2):
                for col in range(0, width):
                    if not Puzzle.is_empty_floor(strings[row][col]):
                        cell_found = True
                        if col > x or (col >= x and row > y):
                            x = col
                            y = row

            if cell_found:
                return index_1d(x, y, width)

            return None

        normalized = HexobanPuzzleParser.normalize_width(strings)
        width = len(normalized[0]) if normalized else 0
        height = len(normalized)

        if width == 0 or height == 0:
            return None

        rightmost_in_even_rows = rightmost_finder(normalized, 0)
        rightmost_in_odd_rows = rightmost_finder(normalized, 1)

        if rightmost_in_even_rows is None or rightmost_in_odd_rows is None:
            return index_1d(0, 0, width)

        odd_x = X(rightmost_in_odd_rows, width)
        odd_y = Y(rightmost_in_odd_rows, width)
        even_x = X(rightmost_in_even_rows, width)
        even_y = Y(rightmost_in_even_rows, width)

        if odd_x > even_x:
            return rightmost_in_odd_rows
        elif even_x > odd_x:
            return rightmost_in_even_rows

        return rightmost_in_odd_rows if odd_y > even_y else rightmost_in_even_rows

    def convert_to_internal(self, board: str) -> Tuple[List[str], bool]:
        # Converts textual Hexoban into 2D array and validates textual layout
        (
            parsed,
            width,
            height,
            even_row_x_parity,
            odd_row_x_parity,
        ) = self._preparse_board(board)

        # Handle empty board
        if width == 0 or height == 0:
            return [], True

        elif even_row_x_parity < 0 or odd_row_x_parity < 0:
            internal = height * [int(width / 2) * Puzzle.VISIBLE_FLOOR]
            return internal, True

        layout_ok = True
        internal: List[str] = []

        for y in range(0, height):
            if not layout_ok:
                break
            internal_line = ""
            for x in range(0, width):
                if not layout_ok:
                    break

                layout_ok, should_copy_cell = self._analyze_text_cell_position(
                    parsed[y][x], x, y, odd_row_x_parity, even_row_x_parity
                )

                if layout_ok and should_copy_cell:
                    internal_line += parsed[y][x]

            if layout_ok:
                internal.append(internal_line)

        if layout_ok:
            internal = PuzzleParser.normalize_width(internal, Puzzle.VISIBLE_FLOOR)

        return internal, layout_ok

    def _is_type2(self, strings: List[str]) -> bool:
        return not self.is_type1(strings)

    @staticmethod
    def _analyze_text_cell_position(
        cell: str, x: int, y: int, odd_row_x_parity: int, even_row_x_parity: int
    ) -> Tuple[bool, bool]:
        y_parity = y % 2
        x_parity = x % 2

        layout_ok = True
        # Is current cell part of board or only part of textual layout?
        is_cell_for_layout = False
        should_copy_cell = False

        # Check if textual encoding (layout) is legal. Positions of all
        # board elements in textual layout depend on position of first
        # non floor element. If that element is (odd column, even row)
        # than all other element in even rows must be in odd columns.
        # Other cells in even rows must be empty cells and their purpose
        # is to define board textual layout (they are not board elements).
        if y_parity == 0:  # even rows
            if x_parity == even_row_x_parity:
                is_cell_for_layout = False
            else:  # Cell is part of layout, it must be empty
                layout_ok = Puzzle.is_empty_floor(cell)
                is_cell_for_layout = True
        else:  # odd rows
            if x == 0:  # row start half hexes are always layout cells
                layout_ok = Puzzle.is_empty_floor(cell)
                is_cell_for_layout = True
            else:
                if x_parity == odd_row_x_parity:
                    is_cell_for_layout = False
                else:  # Cell is part of layout, it must be empty
                    layout_ok = Puzzle.is_empty_floor(cell)
                    is_cell_for_layout = True

        should_copy_cell = layout_ok and not is_cell_for_layout

        return layout_ok, should_copy_cell

    def _preparse_board(self, board: str) -> Tuple[List[str], int, int, int, int]:
        from ..game import X, Y

        parsed: List[str] = PuzzleParser.cleaned_board_lines(board)
        width = len(parsed[0]) if parsed else 0
        height = len(parsed)

        even_row_x_parity = odd_row_x_parity = -1

        if height != 0 and width != 0:
            # Compensate for scheme2
            has_non_floor_left_in_odd_row = False

            for i in range(0, height):
                if has_non_floor_left_in_odd_row:
                    break
                has_non_floor_left_in_odd_row = has_non_floor_left_in_odd_row or (
                    i % 2 == 1 and not Puzzle.is_empty_floor(parsed[i][0])
                )

            if has_non_floor_left_in_odd_row:
                parsed = self.add_column_left(parsed)
                width += 1

            # Calculate parities
            first_cell = self._find_first_non_floor(parsed)
            if first_cell is not None:
                first_cell_x_parity = X(first_cell, width) % 2
                first_cell_y_parity = Y(first_cell, width) % 2

                if first_cell_y_parity == 0:
                    even_row_x_parity = first_cell_x_parity
                else:
                    even_row_x_parity = (first_cell_x_parity + 1) % 2

                odd_row_x_parity = (even_row_x_parity + 1) % 2

        return parsed, width, height, even_row_x_parity, odd_row_x_parity

    @staticmethod
    def add_column_left(strings: List[str]) -> List[str]:
        return [Puzzle.VISIBLE_FLOOR + line for line in strings]

    @staticmethod
    def add_column_right(strings: List[str]) -> List[str]:
        return [line + Puzzle.VISIBLE_FLOOR for line in strings]

    @staticmethod
    def add_row_top(strings: List[str]) -> List[str]:
        w = HexobanPuzzleParser.calculate_width(strings)
        return [w * Puzzle.VISIBLE_FLOOR] + strings

    @staticmethod
    def remove_column_right(strings: List[str]) -> List[str]:
        return [line[:-1] for line in strings]

    @staticmethod
    def reverse_columns(strings: List[str]) -> List[str]:
        return ["".join(c for c in reversed(line)) for line in strings]

    @staticmethod
    def remove_row_top(strings: List[str]) -> List[str]:
        return strings[1:]

    @staticmethod
    def remove_row_bottom(strings: List[str]) -> List[str]:
        return strings[:-1]

    @staticmethod
    def _find_first_non_floor(strings: List[str]):
        from ..game import index_1d

        normalized = HexobanPuzzleParser.normalize_width(strings)
        width = len(normalized[0]) if normalized else 0
        height = len(normalized)

        if width == 0 or height == 0:
            return None

        x = y = 0
        non_floor_found = False

        for row in range(0, height):
            if non_floor_found:
                break
            for column in range(0, width):
                if non_floor_found:
                    break

                if not Puzzle.is_empty_floor(normalized[row][column]) and (
                    column > x or row > y
                ):
                    x = column
                    y = row
                    non_floor_found = True

        if non_floor_found:
            return index_1d(x, y, width)

        # Empty board, assuming scheme1
        return index_1d(1, 0, width)


class HexobanPuzzlePrinter(PuzzlePrinter):
    def print(
        self,
        parsed_board: List[str],
        width: int,
        height: int,
        use_visible_floor: bool,
        rle_encode: bool,
    ) -> str:
        retv = HexobanTextConverter().convert_to_string(
            parsed_board, width, height, use_visible_floor
        )
        if rle_encode:
            return Rle.encode(retv)
        return retv


class HexobanPuzzleParser(PuzzleParser):
    def parse(self, board: str) -> List[str]:
        parsed, layout_ok = HexobanTextConverter().convert_to_internal(board)

        if layout_ok:
            return parsed

        raise ValueError(
            "String can't be parsed to HexobanPuzzle. Probable cause is invalid "
            "text layout meaning either missing or misaligned filler floor "
            "characters."
        )


class HexobanPuzzleResizer(PuzzleResizer):
    def reverse_columns(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        converter = HexobanTextConverter()

        printed_board = converter.convert_to_string(
            parsed_board, width, height, True
        ).splitlines()

        if converter.is_type1(printed_board):
            printed_board = converter.add_column_left(printed_board)
        else:
            printed_board = converter.add_column_right(printed_board)

        printed_board = converter.reverse_columns(printed_board)
        printed_board = converter.remove_column_right(printed_board)

        new_parsed_board, layout_ok = converter.convert_to_internal(
            "\n".join(printed_board)
        )
        new_width = len(new_parsed_board[0]) if new_parsed_board else 0
        new_height = len(new_parsed_board)

        return list("".join(new_parsed_board)), new_width, new_height

    def add_row_top(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        converter = HexobanTextConverter()

        printed_board = converter.convert_to_string(
            parsed_board, width, height, True
        ).splitlines()

        printed_board = converter.add_row_top(printed_board)

        new_parsed_board, layout_ok = converter.convert_to_internal(
            "\n".join(printed_board)
        )
        new_width = len(new_parsed_board[0]) if new_parsed_board else 0
        new_height = len(new_parsed_board)

        return list("".join(new_parsed_board)), new_width, new_height

    def remove_row_top(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        converter = HexobanTextConverter()

        printed_board = converter.convert_to_string(
            parsed_board, width, height, True
        ).splitlines()

        printed_board = converter.remove_row_top(printed_board)

        new_parsed_board, layout_ok = converter.convert_to_internal(
            "\n".join(printed_board)
        )
        new_width = len(new_parsed_board[0]) if new_parsed_board else 0
        new_height = len(new_parsed_board)

        return list("".join(new_parsed_board)), new_width, new_height

    def remove_row_bottom(
        self, parsed_board: List[str], width: int, height: int
    ) -> Tuple[List[str], int, int]:
        converter = HexobanTextConverter()

        printed_board = converter.convert_to_string(
            parsed_board, width, height, True
        ).splitlines()

        printed_board = converter.remove_row_bottom(printed_board)

        new_parsed_board, layout_ok = converter.convert_to_internal(
            "\n".join(printed_board)
        )
        new_width = len(new_parsed_board[0]) if new_parsed_board else 0
        new_height = len(new_parsed_board)

        return list("".join(new_parsed_board)), new_width, new_height



"""
Module that handles Sokoban gameplay.
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base_tessellation import BaseTessellation, Tessellation
    from .board_cell import BoardCell
    from .board_graph import BoardGraph
    from .board_manager import (
        BoardManager,
        BoxGoalSwitchError,
        CellAlreadyOccupiedError,
    )
    from .board_state import BoardState
    from .config import Config, Direction, GraphType
    from .hashed_board_manager import HashedBoardManager
    from .hexoban_tessellation import HexobanTessellation
    from .mover import IllegalMoveError, Mover, NonPlayableBoardError, SolvingMode
    from .octoban_tessellation import OctobanTessellation
    from .pusher_step import PusherStep
    from .sokoban_plus import SokobanPlus, SokobanPlusDataError
    from .sokoban_tessellation import SokobanTessellation
    from .trioban_tessellation import TriobanTessellation


else:
    try:
        from sokoenginepyext.game import (
            BaseTessellation,
            BoardCell,
            BoardGraph,
            BoardManager,
            BoardState,
            BoxGoalSwitchError,
            CellAlreadyOccupiedError,
            Config,
            Direction,
            GraphType,
            HashedBoardManager,
            HexobanTessellation,
            IllegalMoveError,
            Mover,
            NonPlayableBoardError,
            OctobanTessellation,
            PusherStep,
            SokobanPlus,
            SokobanPlusDataError,
            SokobanTessellation,
            SolvingMode,
            Tessellation,
            TriobanTessellation,
        )

    except ImportError:
        from .base_tessellation import BaseTessellation, Tessellation
        from .board_cell import BoardCell
        from .board_graph import BoardGraph
        from .board_manager import (
            BoardManager,
            BoxGoalSwitchError,
            CellAlreadyOccupiedError,
        )
        from .board_state import BoardState
        from .config import Config, Direction, GraphType
        from .hashed_board_manager import HashedBoardManager
        from .hexoban_tessellation import HexobanTessellation
        from .mover import IllegalMoveError, Mover, NonPlayableBoardError, SolvingMode
        from .octoban_tessellation import OctobanTessellation
        from .pusher_step import PusherStep
        from .sokoban_plus import SokobanPlus, SokobanPlusDataError
        from .sokoban_tessellation import SokobanTessellation
        from .trioban_tessellation import TriobanTessellation

from .base_tessellation import (
    COLUMN,
    ROW,
    X,
    Y,
    index_1d,
    is_on_board_1d,
    is_on_board_2d,
)
from .mover_commands import JumpCommand, MoveCommand, SelectPusherCommand



from pathlib import Path

from sokoenginepy.io.utilities import in_batches

_SELF_DIR = Path(__file__).absolute().resolve().parent


def encode_spec_to_C_header():
    src_dir = (_SELF_DIR / ".." / "src").resolve()
    src_path = src_dir / "sokoenginepy" / "io" / "SOK_format_specification.txt"
    dst_path = (
        src_dir / "libsokoengine" / "io" / "collection" / "SOK_format_specification.h"
    )

    print(f"Generating \n    {dst_path}\nfrom\n    {src_path}")

    with open(src_path, "r") as f:
        data = "\n".join(_.strip() for _ in f.readlines())
        data = data.encode("UTF-8")

    with open(dst_path, "w") as f:
        f.write("static const std::string SOK_format_specification = {\n")

        lines = [
            "  " + ", ".join(batch)
            for batch in in_batches((f"0x{b:02x}" for b in data), of_size=12)
        ]

        f.write(",\n".join(lines))

        f.write("\n};\n")


if __name__ == "__main__":
    encode_spec_to_C_header()



from __future__ import annotations

from typing import Dict, Final, Optional, Tuple

from ..io import CellOrientation, Snapshot
from .base_tessellation import COLUMN, ROW, BaseTessellation, index_1d, is_on_board_2d
from .config import Direction
from .utilities import inverted


class OctobanTessellation(BaseTessellation):
    """
    Board space is laid out on alternating squares and octagons with origin of
    coordinate system being octagon. Tessellation allows all 8 directions of movement
    from Direction and depending on current pusher position some of these directions do
    not result in successful move.

    Direction <-> character mapping:

    ====  ==========  =====  ==========  ====  ==========  ====  ==========
    UP    NORTH_EAST  RIGHT  SOUTH_EAST  DOWN  SOUTH_WEST  LEFT  NORTH_WEST
    ====  ==========  =====  ==========  ====  ==========  ====  ==========
    u, U  n, N        r, R   e, E        d, D  s, S        l, L  w, W
    ====  ==========  =====  ==========  ====  ==========  ====  ==========
    """

    _LEGAL_DIRECTIONS: Final[Tuple[Direction, ...]] = (
        Direction.LEFT,
        Direction.RIGHT,
        Direction.UP,
        Direction.DOWN,
        Direction.NORTH_EAST,
        Direction.NORTH_WEST,
        Direction.SOUTH_EAST,
        Direction.SOUTH_WEST,
    )

    _CHR_TO_PUSHER_STEP: Final[Dict[str, Tuple[Direction, bool]]] = {
        Snapshot.l: (Direction.LEFT, False),
        Snapshot.L: (Direction.LEFT, True),
        Snapshot.r: (Direction.RIGHT, False),
        Snapshot.R: (Direction.RIGHT, True),
        Snapshot.u: (Direction.UP, False),
        Snapshot.U: (Direction.UP, True),
        Snapshot.d: (Direction.DOWN, False),
        Snapshot.D: (Direction.DOWN, True),
        Snapshot.w: (Direction.NORTH_WEST, False),
        Snapshot.W: (Direction.NORTH_WEST, True),
        Snapshot.e: (Direction.SOUTH_EAST, False),
        Snapshot.E: (Direction.SOUTH_EAST, True),
        Snapshot.n: (Direction.NORTH_EAST, False),
        Snapshot.N: (Direction.NORTH_EAST, True),
        Snapshot.s: (Direction.SOUTH_WEST, False),
        Snapshot.S: (Direction.SOUTH_WEST, True),
    }

    _PUSHER_STEP_TO_CHR: Final[Dict[Tuple[Direction, bool], str]] = inverted(
        _CHR_TO_PUSHER_STEP
    )

    _NEIGHBOR_SHIFT: Final[Dict[Direction, Tuple[int, int]]] = {
        Direction.LEFT: (0, -1),
        Direction.RIGHT: (0, 1),
        Direction.UP: (-1, 0),
        Direction.DOWN: (1, 0),
        Direction.NORTH_WEST: (-1, -1),
        Direction.NORTH_EAST: (-1, 1),
        Direction.SOUTH_WEST: (1, -1),
        Direction.SOUTH_EAST: (1, 1),
    }

    def neighbor_position(
        self, position: int, direction: Direction, board_width: int, board_height: int
    ) -> Optional[int]:
        if self.cell_orientation(
            position, board_width, board_height
        ) != CellOrientation.OCTAGON and (
            direction == Direction.NORTH_EAST
            or direction == Direction.NORTH_WEST
            or direction == Direction.SOUTH_EAST
            or direction == Direction.SOUTH_WEST
        ):
            return None

        row = ROW(position, board_width)
        column = COLUMN(position, board_width)
        row_shift, column_shift = self._NEIGHBOR_SHIFT.get(direction, (None, None))

        if row_shift is None or column_shift is None:
            raise ValueError(direction)

        row += row_shift
        column += column_shift

        if is_on_board_2d(column, row, board_width, board_height):
            return index_1d(column, row, board_width)

        return None

    def cell_orientation(
        self, position: int, board_width: int, board_height: int
    ) -> CellOrientation:
        row = ROW(position, board_width)
        column = COLUMN(position, board_width)
        return (
            CellOrientation.OCTAGON
            if (column + (row % 2)) % 2 == 0
            else CellOrientation.DEFAULT
        )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

from .puzzle import Puzzle

if TYPE_CHECKING:
    from ..game import Tessellation


class Collection:
    """
    Collection of one or more game puzzles.

    Attributes:
        title(str): Title
        author(str): Author
    """

    def __init__(
        self,
        title: str = "",
        author: str = "",
        created_at: str = "",
        updated_at: str = "",
        notes: Optional[List[str]] = None,
    ):
        self.title = title
        self.author = author
        self.created_at = created_at
        self.updated_at = updated_at
        self.notes: List[str] = notes or []
        self.puzzles: List[Puzzle] = []

    @staticmethod
    def _extension_to_tessellation_hint(path: Union[str, Path]) -> Tessellation:
        from ..game import Tessellation

        file_extension = Path(path).suffix
        if file_extension == ".tsb":
            return Tessellation.TRIOBAN
        elif file_extension == ".hsb":
            return Tessellation.HEXOBAN
        else:
            return Tessellation.SOKOBAN

    def load(
        self, path: Union[str, Path], tessellation_hint: Optional[Tessellation] = None
    ):
        from .sok_file_format import SOKFileFormat

        with open(path, "r") as f:
            SOKFileFormat.read(
                f, self, tessellation_hint or self._extension_to_tessellation_hint(path)
            )

    def save(self, path: Union[str, Path]):
        from .sok_file_format import SOKFileFormat

        with open(path, "w") as f:
            SOKFileFormat.write(self, f)



from __future__ import annotations

from typing import Final, Mapping, Optional, Tuple

from ..io import CellOrientation, Snapshot
from .base_tessellation import (
    COLUMN,
    ROW,
    BaseTessellation,
    Direction,
    GraphType,
    index_1d,
    is_on_board_2d,
)
from .config import Direction, GraphType
from .utilities import inverted


class TriobanTessellation(BaseTessellation):
    """
    Board is laid out on alternating triangles with origin triangle pointing down.

    Direction <-> character mapping:

    ====  =====  ==========  ==========  ==========  ==========
    LEFT  RIGHT  NORTH_EAST  NORTH_WEST  SOUTH_EAST  SOUTH_WEST
    ====  =====  ==========  ==========  ==========  ==========
    l, L  r, R   n, N        u, U        d, D        s, S
    ====  =====  ==========  ==========  ==========  ==========

    Depending on pusher position, not all move directions are allowed on all board
    positions:

    .. image:: /images/trioban_am.png
        :alt: Trioban movement
    """

    _LEGAL_DIRECTIONS: Final[Tuple[Direction, ...]] = (
        Direction.LEFT,
        Direction.RIGHT,
        Direction.NORTH_EAST,
        Direction.NORTH_WEST,
        Direction.SOUTH_EAST,
        Direction.SOUTH_WEST,
    )

    _CHR_TO_PUSHER_STEP: Final[Mapping[str, Tuple[Direction, bool]]] = {
        Snapshot.l: (Direction.LEFT, False),
        Snapshot.L: (Direction.LEFT, True),
        Snapshot.r: (Direction.RIGHT, False),
        Snapshot.R: (Direction.RIGHT, True),
        Snapshot.n: (Direction.NORTH_EAST, False),
        Snapshot.N: (Direction.NORTH_EAST, True),
        Snapshot.u: (Direction.NORTH_WEST, False),
        Snapshot.U: (Direction.NORTH_WEST, True),
        Snapshot.d: (Direction.SOUTH_EAST, False),
        Snapshot.D: (Direction.SOUTH_EAST, True),
        Snapshot.s: (Direction.SOUTH_WEST, False),
        Snapshot.S: (Direction.SOUTH_WEST, True),
    }

    _PUSHER_STEP_TO_CHR: Final[Mapping[Tuple[Direction, bool], str]] = inverted(
        _CHR_TO_PUSHER_STEP
    )

    @property
    def graph_type(self) -> GraphType:
        return GraphType.DIRECTED_MULTI

    def neighbor_position(
        self, position: int, direction: Direction, board_width: int, board_height: int
    ) -> Optional[int]:
        row = ROW(position, board_width)
        column = COLUMN(position, board_width)
        triangle_points_down = (
            self.cell_orientation(position, board_width, board_height)
            == CellOrientation.TRIANGLE_DOWN
        )

        dx, dy = 0, 0
        if direction == Direction.LEFT:
            dy = 0
            dx = -1
        elif direction == Direction.RIGHT:
            dy = 0
            dx = 1
        elif direction == Direction.NORTH_EAST:
            if triangle_points_down:
                dy = -1
                dx = 0
            else:
                dy = 0
                dx = 1
        elif direction == Direction.NORTH_WEST:
            if triangle_points_down:
                dy = -1
                dx = 0
            else:
                dy = 0
                dx = -1
        elif direction == Direction.SOUTH_EAST:
            if triangle_points_down:
                dy = 0
                dx = 1
            else:
                dy = 1
                dx = 0
        elif direction == Direction.SOUTH_WEST:
            if triangle_points_down:
                dy = 0
                dx = -1
            else:
                dy = 1
                dx = 0
        else:
            raise ValueError(direction)

        row += dy
        column += dx

        if is_on_board_2d(column, row, board_width, board_height):
            return index_1d(column, row, board_width)

        return None

    def cell_orientation(
        self, position: int, board_width: int, board_height: int
    ) -> CellOrientation:
        row = ROW(position, board_width)
        column = COLUMN(position, board_width)
        return (
            CellOrientation.TRIANGLE_DOWN
            if (column + (row % 2)) % 2 == 0
            else CellOrientation.DEFAULT
        )



from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Final, List, Set, Union

import lark

from .rle import Rle
from .snapshot import Snapshot
from .utilities import is_blank

if TYPE_CHECKING:
    from ..game import BaseTessellation, PusherStep, Tessellation


class Constants:
    l: Final[str] = Snapshot.l
    u: Final[str] = Snapshot.u
    r: Final[str] = Snapshot.r
    d: Final[str] = Snapshot.d
    L: Final[str] = Snapshot.L
    U: Final[str] = Snapshot.U
    R: Final[str] = Snapshot.R
    D: Final[str] = Snapshot.D
    w: Final[str] = Snapshot.w
    W: Final[str] = Snapshot.W
    e: Final[str] = Snapshot.e
    E: Final[str] = Snapshot.E
    n: Final[str] = Snapshot.n
    N: Final[str] = Snapshot.N
    s: Final[str] = Snapshot.s
    S: Final[str] = Snapshot.S

    JUMP_BEGIN: Final[str] = Snapshot.JUMP_BEGIN
    JUMP_END: Final[str] = Snapshot.JUMP_END
    PUSHER_CHANGE_BEGIN: Final[str] = Snapshot.PUSHER_CHANGE_BEGIN
    PUSHER_CHANGE_END: Final[str] = Snapshot.PUSHER_CHANGE_END
    CURRENT_POSITION_CH: Final[str] = Snapshot.CURRENT_POSITION_CH

    MOVE_CHARACTERS: Final[Set[str]] = {l, u, r, d, n, s, e, w}
    PUSH_CHARACTERS: Final[Set[str]] = {L, U, R, D, N, S, E, W}
    MARKERS: Final[Set[str]] = {
        JUMP_BEGIN,
        JUMP_END,
        PUSHER_CHANGE_BEGIN,
        PUSHER_CHANGE_END,
        CURRENT_POSITION_CH,
    }

    RE_SNAPSHOT_STRING = re.compile(
        r"^([0-9\s"
        + re.escape("".join(MOVE_CHARACTERS))
        + re.escape("".join(PUSH_CHARACTERS))
        + re.escape("".join(MARKERS))
        + re.escape("".join({Rle.GROUP_START, Rle.GROUP_END, Rle.EOL}))
        + "])*$"
    )


@dataclass
class Steps:
    data: str

    @property
    def pushes_count(self):
        return sum(1 for _ in self.data if _ in Constants.PUSH_CHARACTERS)

    @property
    def moves_count(self):
        return sum(1 for _ in self.data if _ in Constants.MOVE_CHARACTERS)

    def __str__(self):
        return self.data

    @classmethod
    def converted(cls, data: str, tessellation: BaseTessellation) -> List[PusherStep]:
        retv: List[PusherStep] = []
        for _ in data:
            if _ == Constants.CURRENT_POSITION_CH:
                if retv:
                    retv[-1].is_current_pos = True
            else:
                step = tessellation.char_to_pusher_step(_)
                retv.append(step)

        return retv

    def pusher_steps(self, tessellation: BaseTessellation) -> List[PusherStep]:
        return self.converted(self.data, tessellation)


@dataclass
class Jump:
    data: str
    pushes_count: int = field(init=False, repr=False, default=0)

    @property
    def moves_count(self):
        return sum(1 for _ in self.data if _ in Constants.MOVE_CHARACTERS)

    def __str__(self):
        return Constants.JUMP_BEGIN + self.data + Constants.JUMP_END

    def pusher_steps(self, tessellation: BaseTessellation) -> List[PusherStep]:
        retv = Steps.converted(self.data, tessellation)
        for _ in retv:
            _.is_jump = True
        return retv


@dataclass
class PusherSelection:
    data: str
    pushes_count: int = field(init=False, repr=False, default=0)
    moves_count: int = field(init=False, repr=False, default=0)

    def __str__(self):
        return Constants.PUSHER_CHANGE_BEGIN + self.data + Constants.PUSHER_CHANGE_END

    def pusher_steps(self, tessellation: BaseTessellation) -> List[PusherStep]:
        retv = Steps.converted(self.data, tessellation)
        for _ in retv:
            _.is_pusher_selection = True
        return retv


MovementTokens = List[Union[Jump, PusherSelection, Steps]]


class Parser:
    GRAMMAR = f"""
        snapshot: (jump | pusher_selection | steps)+

        jump: jump_begin moves* jump_end
        pusher_selection: pusher_change_begin moves pusher_change_end
        steps: (moves | pushes)+

        jump_begin: "{Constants.JUMP_BEGIN}"
        jump_end: "{Constants.JUMP_END}"
        pusher_change_begin: "{Constants.PUSHER_CHANGE_BEGIN}"
        pusher_change_end: "{Constants.PUSHER_CHANGE_END}"

        moves: /[{''.join(Constants.MOVE_CHARACTERS)}\\{Constants.CURRENT_POSITION_CH}]+/
        pushes: /[{''.join(Constants.PUSH_CHARACTERS)}]+/

        %import common.WS
        %ignore WS
    """

    # Reference: .venv/lib/python3.9/site-packages/lark/grammars/common.lark
    #
    #     WS: /[ \t\f\r\n]/+

    PARSER = lark.Lark(GRAMMAR, parser="lalr", start="snapshot")

    @classmethod
    def parse(cls, data: str) -> MovementTokens:
        if is_blank(data):
            return []

        data = Rle.decode(data)

        try:
            parsed = LarkTreeTransformer().transform(cls.PARSER.parse(data))

        except lark.exceptions.VisitError as e:
            # raise ValueError(str(e))
            raise e.orig_exc from e

        except lark.exceptions.UnexpectedInput as e:
            raise ValueError("Unexpected input in Snapshot string! " + str(e)) from e

        return parsed


class LarkTreeTransformer(lark.Transformer):
    def snapshot(self, args: MovementTokens):
        return args

    def jump(self, args):
        return Jump("".join(args))

    def pusher_selection(self, args):
        return PusherSelection("".join(args))

    def steps(self, args):
        return Steps("".join(args))

    def pushes(self, args):
        return "".join(args)

    def moves(self, args):
        return "".join(args)

    def jump_begin(self, args):
        return lark.visitors.Discard

    def jump_end(self, args):
        return lark.visitors.Discard

    def pusher_change_begin(self, args):
        return lark.visitors.Discard

    def pusher_change_end(self, args):
        return lark.visitors.Discard



from __future__ import annotations

from typing import Dict, Final, Optional, Tuple

from ..io import Snapshot
from .base_tessellation import COLUMN, ROW, BaseTessellation, index_1d, is_on_board_2d
from .config import Direction
from .utilities import inverted


class HexobanTessellation(BaseTessellation):

    """
    Board space is laid out on vertical hexagons with following coordinate system:

    .. image:: /images/hexoban_coordinates.png
        :alt: Hexoban coordinates

    Textual representation uses two characters for each hexagon. This allows different
    encoding schemes.

    .. |img1| image:: /images/hexoban_scheme1.png
    .. |img2| image:: /images/hexoban_scheme2.png

    +----------+----------+
    | Scheme 1 | Scheme 2 |
    +==========+==========+
    |  |img1|  |  |img2|  |
    +----------+----------+

    As long as encoding of single board is consistent, all methods handle any scheme
    transparently - parsing of board strings 'Just Works (TM)'

    Direction <-> character mapping:

    ====  =====  ==========  ==========  ==========  ==========
    LEFT  RIGHT  NORTH_WEST  SOUTH_WEST  NORTH_EAST  SOUTH_EAST
    ====  =====  ==========  ==========  ==========  ==========
    l, L  r, R   u, U        d, D        n, N        s, S
    ====  =====  ==========  ==========  ==========  ==========
    """

    _LEGAL_DIRECTIONS: Final[Tuple[Direction, ...]] = (
        Direction.LEFT,
        Direction.RIGHT,
        Direction.NORTH_EAST,
        Direction.NORTH_WEST,
        Direction.SOUTH_EAST,
        Direction.SOUTH_WEST,
    )

    _CHR_TO_PUSHER_STEP: Final[Dict[str, Tuple[Direction, bool]]] = {
        Snapshot.l: (Direction.LEFT, False),
        Snapshot.L: (Direction.LEFT, True),
        Snapshot.r: (Direction.RIGHT, False),
        Snapshot.R: (Direction.RIGHT, True),
        Snapshot.u: (Direction.NORTH_WEST, False),
        Snapshot.U: (Direction.NORTH_WEST, True),
        Snapshot.d: (Direction.SOUTH_EAST, False),
        Snapshot.D: (Direction.SOUTH_EAST, True),
        Snapshot.n: (Direction.NORTH_EAST, False),
        Snapshot.N: (Direction.NORTH_EAST, True),
        Snapshot.s: (Direction.SOUTH_WEST, False),
        Snapshot.S: (Direction.SOUTH_WEST, True),
    }

    _PUSHER_STEP_TO_CHR: Final[Dict[Tuple[Direction, bool], str]] = inverted(
        _CHR_TO_PUSHER_STEP
    )

    def neighbor_position(
        self, position: int, direction: Direction, board_width: int, board_height: int
    ) -> Optional[int]:
        row = ROW(position, board_width)
        column = COLUMN(position, board_width)

        if direction == Direction.LEFT:
            column -= 1
        elif direction == Direction.RIGHT:
            column += 1
        elif direction == Direction.NORTH_EAST:
            column += row % 2
            row -= 1
        elif direction == Direction.NORTH_WEST:
            column -= (row + 1) % 2
            row -= 1
        elif direction == Direction.SOUTH_EAST:
            column += row % 2
            row += 1
        elif direction == Direction.SOUTH_WEST:
            column -= (row + 1) % 2
            row += 1
        else:
            raise ValueError(direction)

        if is_on_board_2d(column, row, board_width, board_height):
            return index_1d(column, row, board_width)
        return None



from __future__ import annotations

from itertools import groupby
from typing import Final, Optional, Set

import lark


class Rle:
    """
    Rle encoding and decoding.
    """

    GROUP_START: Final[str] = "("
    GROUP_END: Final[str] = ")"
    EOL: Final[str] = "|"

    @classmethod
    def encode(cls, line: str) -> str:
        """RLE encodes string, ie "aaabbbb" becomes "3a4b"."""
        if not line:
            return line

        digits = {str(_) for _ in range(10)}
        if all(_ in digits for _ in line):
            raise ValueError("Cant encode fully numeric strings!")

        encoded = [(len(list(g)), k) for k, g in groupby(line.replace("\n", cls.EOL))]
        return "".join("".join((str(c) if c > 1 else "", v)) for c, v in encoded)

    @classmethod
    def decode(cls, data: Optional[str]) -> str:
        """
        Decodes RLE encoded string.

        Supports RLE groups, ie strings like "3(a2b)4b"
        """
        if not data:
            return ""

        try:
            parsed = LarkTreeTransformer().transform(cls._PARSER.parse(data))

        except lark.exceptions.VisitError as e:
            # raise ValueError(str(e))
            raise e.orig_exc from e

        except lark.exceptions.UnexpectedInput as e:
            raise ValueError("Unexpected input in Rle string! " + str(e)) from e

        # Rle should only modify it's own tokens, leaving the rest of input as is. This
        # means all following examples must work like this:
        #
        #     Rle.decode("\n\n\n") -> "\n\n\n"
        #     Rle.decode("\n|\n")  -> "\n\n\n"
        #     Rle.decode("|||")    -> "\n\n\n"
        #
        # To effectively normalize new lines in final output, we must do following:
        return "".join(parsed)

    _GRAMMAR = f"""
        data: expr+
        expr: atoms | term | group
        term: count (ATOM | group)
        group: _GB expr+ _GE
        atoms: ATOM+
        count: DIGITS

        ATOM: /([^0-9\\{GROUP_START}\\{GROUP_END}])|([ \\n\\r\\t\\{EOL}])/
        _GB: "{GROUP_START}"
        _GE: "{GROUP_END}"
        DIGITS: /[0-9]+/
    """

    _PARSER = lark.Lark(_GRAMMAR, parser="lalr", start="data")


class LarkTreeTransformer(lark.Transformer):
    def data(self, lines):
        return list(lines)

    def expr(self, args):
        return "".join(args)

    def term(self, args):
        return args[0] * args[1]

    def group(self, args):
        return "".join(args)

    def atoms(self, args):
        return "".join(args).replace(Rle.EOL, "\n")

    def count(self, args):
        return int(args[0])



import os

import setuptools

# Available at setup time due to pyproject.toml
from pybind11.setup_helpers import ParallelCompile, Pybind11Extension, naive_recompile


class SokoenginepyextOptions:
    """
    Describes ``sokoenginepyext`` native C++ extension for ``sokoenginepy``.

    On ``Linux``, ``pip install sokoenginepy`` will try to configure and build native
    extension. If build fails, ``sokoenginepy`` will be installed without native
    extension. To succeed, boost header have to be in system include path.

    On all other systems, native extension will not be installed.
    """

    NAME = "sokoenginepyext"

    SHOULD_TRY_BUILD = (
        # We support building only on Linux...
        os.name == "posix"
        # ... and not on Read The Docs
        and os.environ.get("READTHEDOCS", "false").lower()
        not in ["yes", "true", "y", "1"]
        # ... and allow build to be controlled by SOKOENGINEPYEXT_BUILD
        # environment variable
        and os.environ.get("SOKOENGINEPYEXT_BUILD", "true").lower()
        in ["yes", "true", "y", "1"]
    )

    IS_DEBUG = os.environ.get("SOKOENGINEPYEXT_DEBUG", "false").lower() in [
        "yes",
        "true",
        "y",
        "1",
    ]

    CXXFLAGS_RELEASE = ["-O3", "-flto", "-UDEBUG", "-DNDEBUG"]
    CXXFLAGS_DEBUG = ["-g3", "-O0", "-UNDEBUG", "-DDEBUG"]
    CXXFLAGS = [
        "-fPIC",
        "-DBOOST_BIND_NO_PLACEHOLDERS",
        "-DBOOST_MULTI_INDEX_DISABLE_SERIALIZATION",
        "-DLIBSOKOENGINE_DLL",
    ] + (CXXFLAGS_DEBUG if IS_DEBUG else CXXFLAGS_RELEASE)

    LDFLAGS = ["-flto"] if not IS_DEBUG else []

    SOURCES = sorted(
        [
            os.path.join(dir_path, file_name)
            for _ in ["src/libsokoengine", "src/sokoenginepyext"]
            for dir_path, directories, files in os.walk(_)
            for file_name in files
            if (
                file_name.endswith(".cpp")
                and file_name not in {"playground.cpp", "benchmarks.cpp"}
            )
        ]
    )

    INCLUDE_DIRS = ["/tmp/cmake_cache"] + [
        "src/libsokoengine",
        "src/libsokoengine/game",
        "src/libsokoengine/io",
        "src/sokoenginepyext",
    ]

    CPPITERTOOLS_DIR = os.path.abspath("/tmp/cmake_cache/cppitertools")

    @classmethod
    def fetch_cppitertools(cls):
        if not os.path.exists(cls.CPPITERTOOLS_DIR):
            print("Cloning cppitertools...")
            os.system(
                'git clone --branch v1.0 https://github.com/ryanhaining/cppitertools.git "{}"'.format(
                    cls.CPPITERTOOLS_DIR
                )
            )
        return True


if SokoenginepyextOptions.SHOULD_TRY_BUILD:
    SokoenginepyextOptions.fetch_cppitertools()
    ParallelCompile(
        "SOKOENGINEPYEXT_NUM_BUILD_JOBS", needs_recompile=naive_recompile
    ).install()
    ext = Pybind11Extension(
        name=SokoenginepyextOptions.NAME,
        sources=SokoenginepyextOptions.SOURCES,
        include_dirs=SokoenginepyextOptions.INCLUDE_DIRS,
        optional=True,
        cxx_std=17,
        extra_compile_args=SokoenginepyextOptions.CXXFLAGS,
        extra_link_args=SokoenginepyextOptions.LDFLAGS,
        # Example: passing in the version to the compiled code
        # define_macros=[("VERSION_INFO", __version__)],
    )
    setuptools.setup(ext_modules=[ext])

else:
    setuptools.setup()



from __future__ import annotations

import random
from typing import TYPE_CHECKING, List, Optional, Set

from .board_manager import BoardManager
from .board_state import BoardState
from .config import Config

if TYPE_CHECKING:
    from .board_graph import BoardGraph


class HashedBoardManager(BoardManager):
    """:class:`.BoardManager` with Zobrist hashing

    Adds Zobrist hashing on top of :class:`.BoardManager` and keeps it up to date
    when pieces are moved.

    Zobrist hash is 64b integer hash derived from positions of all boxes and pushers
    on board.

    When initialized, :class:`.HashedBoardManager` hashes board using positions and
    IDs of boxes and pushers and produces 64b integer hash. After that, whenever
    position of piece changes, this hash is updated. The ``Zobrist`` part means
    hashing is deterministic which then means that undoing movement will return hash
    value to previous one. All this allows for creation of position tables that
    contain many board layouts and can be quickly compared (since we are not
    comparing positions but only hashes of these positions). Being able to quickly
    compare and find current board layout in some big table, speeds up searching
    through game space which is needed for effective solver implementations.

    Boxes with same Sokoban+ ID are treated as equal meaning that if two of these
    boxes switch position, hash will not change. This also means that hash is
    different when Sokoban+ is enabled from the one when it is disabled

    Pushers are all treated equal, meaning that if two pushers switch position, hash
    will not change

    Notes:
        - enabling/disabling Sokoban+ rehashes the board state
        - changing position of pieces only updates existing hash, it doesn't rehash
          whole board. This means that for example undoing box push would "undo" the
          hash value to the one that was before move was preformed
    """

    def __init__(
        self,
        board: BoardGraph,
        boxorder: Optional[str] = None,
        goalorder: Optional[str] = None,
    ):
        super().__init__(board, boxorder, goalorder)
        self._initial_state_hash = None
        self._state_hash = None
        self._pushers_factors = None
        self._boxes_factors = None
        self._solutions_hashes = None
        self._zobrist_rehash()

    def _zobrist_rehash(self):
        """Recalculates Zobrist hash of board position from scratch."""
        distinct_box_plus_ids = set(
            self.box_plus_id(box_id) for box_id in self.boxes_ids
        )

        # We need random numbers only for this number of positions
        board_without_walls_size = self.board.size - len(self.walls_positions)

        random_pool_size = (
            # one random number for each position pusher can occupy
            board_without_walls_size
            # one random number for position each distinct box can occupy
            + len(distinct_box_plus_ids) * board_without_walls_size
            # one random number for initial hash
            + 1
        )

        # generate required set of random numbers
        # state_backup = random.getstate()
        # Seeding to constant always produces same sequence which then ensures
        # that equal board layouts always produce equal hash
        # random.seed(42)
        random_pool: Set[int] = set()
        while len(random_pool) < random_pool_size:
            random_pool.add(random.getrandbits(64))
        random_pool: List[int] = list(random_pool)
        # random.setstate(state_backup)

        self._initial_state_hash = self._state_hash = random_pool[0]
        random_pool = random_pool[1:]

        # Store position factors for all distinct positions of all distinct
        # boxes
        self._boxes_factors = dict()
        for box_plus_id in distinct_box_plus_ids:
            self._boxes_factors[box_plus_id] = self._insert_wall_zeroes(
                random_pool[:board_without_walls_size]
            )
            random_pool = random_pool[board_without_walls_size:]

        # Store position factors for all distinct pusher positions
        self._pushers_factors = self._insert_wall_zeroes(
            random_pool[:board_without_walls_size]
        )

        for box_id in self.boxes_ids:
            self._state_hash ^= self._boxes_factors[self.box_plus_id(box_id)][
                self.box_position(box_id)
            ]

        for pusher_position in self.pushers_positions.values():
            self._state_hash ^= self._pushers_factors[pusher_position]

    def _insert_wall_zeroes(self, lst):
        src_index = 0

        def choose(position):
            nonlocal src_index
            if position in self.walls_positions:
                return None
            else:
                src_index += 1
                return lst[src_index - 1]

        return [choose(pos) for pos in range(self.board.size)]

    @property
    def state_hash(self) -> int:
        """Zobrist hash of current board state."""
        if self._state_hash is None or self._initial_state_hash is None:
            self._zobrist_rehash()
        return self._state_hash

    @property
    def initial_state_hash(self) -> int:
        """
        Zobrist hash of initial board state (before any movement happened).
        """
        if self._state_hash is None or self._initial_state_hash is None:
            self._zobrist_rehash()
        return self._initial_state_hash

    def external_state_hash(self, board_state) -> Optional[int]:
        """
        Calculates Zobrist hash of given ``board_state`` as if that ``board_state``
        was applied to initial ``board`` (to board where no movement happened).

        ``board_state`` must meet following requirement:

            len(board_state.boxes_positions) == self.boxes_count
            and len(board_state.boxes_positions) == self.goals_count

        Returns:
            Value of hash or None if it can't be calculated
        """
        if (
            len(board_state.boxes_positions) != self.boxes_count
            or len(board_state.boxes_positions) != self.goals_count
        ):
            return None

        retv = self.initial_state_hash
        for index, box_position in enumerate(board_state.boxes_positions):
            retv ^= self._boxes_factors[
                self.box_plus_id(Config.DEFAULT_PIECE_ID + index)
            ][box_position]

        for pusher_position in board_state.pushers_positions:
            retv ^= self._pushers_factors[pusher_position]

        board_state.zobrist_hash = retv

        return retv

    def _box_moved(self, old_position: int, to_new_position: int):
        if old_position != to_new_position:
            box_plus_id = self.box_plus_id(self.box_id_on(to_new_position))
            self._state_hash ^= self._boxes_factors[box_plus_id][old_position]
            self._state_hash ^= self._boxes_factors[box_plus_id][to_new_position]

    def _pusher_moved(self, old_position: int, to_new_position: int):
        if old_position != to_new_position:
            self._state_hash ^= self._pushers_factors[old_position]
            self._state_hash ^= self._pushers_factors[to_new_position]

    @BoardManager.boxorder.setter
    def boxorder(self, rv):
        old_plus_enabled = self.is_sokoban_plus_enabled
        BoardManager.boxorder.fset(self, rv)
        if self.is_sokoban_plus_enabled != old_plus_enabled:
            self._solutions_hashes = None
            self._zobrist_rehash()

    @BoardManager.goalorder.setter
    def goalorder(self, rv):
        old_plus_enabled = self.is_sokoban_plus_enabled
        BoardManager.goalorder.fset(self, rv)
        if self.is_sokoban_plus_enabled != old_plus_enabled:
            self._solutions_hashes = None
            self._zobrist_rehash()

    def enable_sokoban_plus(self):
        old_plus_enabled = self.is_sokoban_plus_enabled
        super().enable_sokoban_plus()
        if self.is_sokoban_plus_enabled != old_plus_enabled:
            self._solutions_hashes = None
            self._zobrist_rehash()

    def disable_sokoban_plus(self):
        old_plus_enabled = self.is_sokoban_plus_enabled
        super().disable_sokoban_plus()
        if self.is_sokoban_plus_enabled != old_plus_enabled:
            self._solutions_hashes = None
            self._zobrist_rehash()

    @property
    def is_solved(self) -> bool:
        return self.state_hash in self.solutions_hashes

    @property
    def solutions_hashes(self) -> Set[int]:
        if not self._solutions_hashes:
            self._solutions_hashes = set(
                h
                for h in (
                    self.external_state_hash(solution) for solution in self.solutions()
                )
                if h
            )
        return self._solutions_hashes

    @property
    def state(self) -> BoardState:
        retv = super().state
        retv.zobrist_hash = self.state_hash
        return retv



import tempfile
import textwrap
from pathlib import Path

import pytest

from sokoenginepy.game import Tessellation
from sokoenginepy.io import Collection, is_blank, snapshot


@pytest.fixture
def input_files_root(resources_root):
    return resources_root / "test_data"


class DescribeCollection:
    def it_loads_regular_collection_file(self, input_files_root):
        collection = Collection()
        collection.load(input_files_root / "Original_and_Extra.sok")

        assert len(collection.puzzles) == 91
        assert len(collection.puzzles[0].snapshots) == 6

        assert collection.title == "Original & Extra"
        assert collection.author == "Thinking Rabbit"
        assert collection.created_at == "2005-03-30  01:05:06"
        assert collection.updated_at == "2012-10-20  09:48:22"
        assert list(collection.notes) == [
            "Originate from: original.slc",
            "",
            "Some collection notes",
        ]

        puzzle = collection.puzzles[0]
        assert puzzle.title == "Level 1"
        assert puzzle.author == "author name"
        assert puzzle.boxorder == "1 2 3 4 5 6"
        assert puzzle.goalorder == "6 5 4 3 2 1"
        assert list(puzzle.notes) == ["Time: 00:01:00", "", "Some puzzle notes"]
        assert not is_blank(puzzle.board)

        expected_board = (
            "\n".join(
                textwrap.dedent(
                    """
                    #####
                    #   #
                    #$  #
                  ###  $##
                  #  $ $ #
                ### # ## #   ######
                #   # ## #####  ..#
                # $  $          ..#
                ##### ### #@##  ..#
                    #     #########
                    #######
                """
                )
                .lstrip("\n")
                .splitlines()
            )
            + "\n"
        )
        assert puzzle.board == expected_board

        snapshot = puzzle.snapshots[0]
        assert list(snapshot.notes) == [
            "Some snapshot notes",
            "Optimizer: YASS 2.86",
            "Time: 00:00:47",
        ]
        expected_moves_data = "".join(
            _.strip()
            for _ in (
                """
                ullluuuLUllDlldddrRRRRRRRRRRRRlllllllllllllulldRRRRRRRRRRRRRdrUlllllll
                uuululldDDuulldddrRRRRRRRRRRuRRlDllllllluuulLulDDDuulldddrRRRRRRRRRRdR
                RlUllllllluuulluuurDDluulDDDDDuulldddrRRRRRRRRRRRllllllluuuLLulDDDuull
                dddrRRRRRRRRRRuRDldR
                """
            ).splitlines()
        )
        assert snapshot.moves_data == expected_moves_data
        assert snapshot.solver == "only human"

    def it_loads_puzzle_without_title(self, input_files_root):
        collection = Collection()
        collection.load(input_files_root / "parser_test_puzzle_no_title.sok")

        assert is_blank(collection.puzzles[0].title)
        assert not is_blank(collection.puzzles[0].board)

    def it_loads_puzzle_with_multiple_snapshots(self, input_files_root):
        collection = Collection()
        collection.load(input_files_root / "parser_test_multiple_snapshots.sok")

        assert len(collection.puzzles[0].snapshots) == 3
        assert collection.puzzles[0].snapshots[1].title == (
            "This is note for first snapshot, but parser "
            + "can't know that and should recognize it as heading of next snapshot."
        )

    def it_loads_last_snapshot_notes(self, input_files_root):
        collection = Collection()
        collection.load(input_files_root / "parser_test_last_snapshot_notes.sok")

        title = "This is note for last snapshot, not heading line"
        note = "but parser can't know that and should recognize it as heading line"

        assert collection.puzzles[1].title == title
        assert list(collection.puzzles[1].notes) == [note]

    def it_correctly_loads_mixed_tessellations_collection(self, input_files_root):
        collection = Collection()
        collection.load(input_files_root / "mixed_collection.sok")

        assert (
            sum(1 for _ in collection.puzzles if _.tessellation == Tessellation.SOKOBAN)
            == 4
        )
        assert (
            sum(1 for _ in collection.puzzles if _.tessellation == Tessellation.HEXOBAN)
            == 5
        )

        for puzzle in collection.puzzles:
            for snapshot in puzzle.snapshots:
                assert puzzle.tessellation == snapshot.tessellation

    class describe_choosing_puzzle_variant:
        def it_defaults_to_sokoban_if_not_specified(self, input_files_root):
            collection = Collection()
            collection.load(
                input_files_root / "parser_test_variant_type_not_specified.sok"
            )

            for puzzle in collection.puzzles:
                assert puzzle.tessellation == Tessellation.SOKOBAN

        def it_uses_value_from_collection_notes_even_if_hint_given(
            self, input_files_root
        ):
            collection = Collection()
            collection.load(
                input_files_root / "parser_test_variant_type_specified_global.sok",
                Tessellation.OCTOBAN,
            )

            for puzzle in collection.puzzles:
                assert puzzle.tessellation == Tessellation.TRIOBAN

        def it_snapshots_get_the_same_value_as_their_puzzle(self, input_files_root):
            collection = Collection()
            collection.load(
                input_files_root / "parser_test_variant_type_specified_global.sok",
                Tessellation.OCTOBAN,
            )

            for puzzle in collection.puzzles:
                for snapshot in puzzle.snapshots:
                    assert snapshot.tessellation == puzzle.tessellation

        def it_uses_value_from_puzzle_notes_and_ignores_hint_and_collection_notes(
            self, input_files_root
        ):
            collection = Collection()
            collection.load(
                input_files_root / "parser_test_variant_type_specified_puzzle1.sok",
                Tessellation.OCTOBAN,
            )

            assert collection.puzzles[0].tessellation == Tessellation.SOKOBAN
            assert collection.puzzles[1].tessellation == Tessellation.HEXOBAN

        def it_uses_value_from_hint_when_nothing_is_specified_in_file(
            self, input_files_root
        ):
            collection = Collection()
            collection.load(
                input_files_root / "parser_test_variant_type_specified_puzzle2.sok",
                Tessellation.TRIOBAN,
            )

            assert collection.puzzles[0].tessellation == Tessellation.TRIOBAN
            assert collection.puzzles[1].tessellation == Tessellation.HEXOBAN

    def it_correctly_writes_collection(self, input_files_root):
        original = Collection()
        original.load(input_files_root / "Original_and_Extra.sok")

        loaded = Collection()
        with tempfile.TemporaryDirectory() as dir_path:
            file_path = Path(dir_path) / "saved_original_and_extra.sok"
            original.save(file_path)
            loaded.load(file_path)

        assert len(original.puzzles) == len(loaded.puzzles)
        for attr in {"title", "author", "created_at", "updated_at", "notes"}:
            assert getattr(original, attr) == getattr(
                loaded, attr
            ), f"Mismatched Collection.{attr}"

        for p_idx in range(len(original.puzzles)):
            original_p = original.puzzles[p_idx]
            loaded_p = loaded.puzzles[p_idx]

            for attr in {
                "title",
                "author",
                "boxorder",
                "goalorder",
                "notes",
                "board",
            }:
                assert getattr(original_p, attr) == getattr(
                    loaded_p, attr
                ), f'Mismatched Puzzle.{attr} for puzzle "{original_p.title}"'

            assert len(original_p.snapshots) == len(loaded_p.snapshots)
            for s_idx in range(len(original_p.snapshots)):
                original_s = original_p.snapshots[s_idx]
                loaded_s = loaded_p.snapshots[s_idx]

                for attr in {"title", "solver", "notes", "moves_data"}:
                    assert getattr(original_s, attr) == getattr(loaded_s, attr), (
                        f"Mismatched Snapshot.{attr} for snapshot {original_p.title}: "
                        f"{original_s.title or s_idx}"
                    )



import pytest
from sokoenginepy.io import Rle


class DescribeRle:
    def it_rle_encodes_string(self):
        assert Rle.encode("aaaabbbbccdeee") == "4a4b2cd3e"
        assert Rle.encode("aaaabbbbccdeeefghij") == "4a4b2cd3efghij"
        assert Rle.encode("aaaabbbbccdeeef") == "4a4b2cd3ef"
        assert Rle.encode("aaaabbfbbccdeee") == "4a2bf2b2cd3e"
        assert Rle.encode("aaaabbfghijbbccdeee") == "4a2bfghij2b2cd3e"
        assert Rle.encode(" aaaaaaaaaa   ") == " 10a3 "
        assert Rle.encode("   ") == "3 "

    def it_rle_line_must_have_at_least_one_non_digit(self):
        with pytest.raises(ValueError):
            assert Rle.encode("1234567890") == "1234567890"

    def it_decodes_rle_string(self):
        assert Rle.decode("4a4b2cd3e") == "aaaabbbbccdeee"
        assert Rle.decode("4a4b2cd3efghij") == "aaaabbbbccdeeefghij"
        assert Rle.decode("4a4b2cd3ef") == "aaaabbbbccdeeef"
        assert Rle.decode("4a2bf2b2cd3e") == "aaaabbfbbccdeee"
        assert Rle.decode("4a2bfghij2b2cd3e") == "aaaabbfghijbbccdeee"
        assert Rle.decode(" 10a3 ") == " aaaaaaaaaa   "
        assert Rle.decode("ccc") == "ccc"
        assert Rle.decode("3a4b") == "aaabbbb"

        assert Rle.decode("3(a2b)4b") == "abbabbabbbbbb"
        assert Rle.decode("aa2(bb)cc2(dd2(ee)ff)") == "aabbbbccddeeeeffddeeeeff"
        assert Rle.decode("2a4b2c2d2e|2e2f2d4e2f") == "aabbbbccddee\neeffddeeeeff"
        assert Rle.decode("2a4b2c2d2e5 2e2f2d4e2f") == "aabbbbccddee     eeffddeeeeff"

    def it_decodes_grouped_rle_string(self):
        assert (
            Rle.decode("2abc3def3(adfdf)2abc3def")
            == "aabcdddefadfdfadfdfadfdfaabcdddef"
        )
        assert (
            Rle.decode("2abc3def3(2a3bc)2abc3def")
            == "aabcdddefaabbbcaabbbcaabbbcaabcdddef"
        )
        assert (
            Rle.decode("2abc3def2(fdsfs2(dfgh)gtr)2abc3def")
            == "aabcdddeffdsfsdfghdfghgtrfdsfsdfghdfghgtraabcdddef"
        )
        assert (
            Rle.decode("2abc3def2(2a2bc2(2a3b)2d)2abc3def")
            == "aabcdddefaabbcaabbbaabbbddaabbcaabbbaabbbddaabcdddef"
        )

        assert Rle.decode("2(abcd)efgh") == "abcdabcdefgh"
        assert Rle.decode("2(3abcd)efgh") == "aaabcdaaabcdefgh"
        assert Rle.decode("2(abcd)3efgh") == "abcdabcdeeefgh"
        assert Rle.decode("(abcd)3efgh") == "abcdeeefgh"
        assert Rle.decode("(ab3cd)efgh") == "abcccdefgh"
        assert Rle.decode("(ab2(3c)d)efgh") == "abccccccdefgh"
        assert Rle.decode("efgh(ab3cd)") == "efghabcccd"

    def it_doesnt_modify_original_newlines_in_any_way(self):
        assert Rle.decode("\n\n\n") == "\n\n\n"
        assert Rle.decode("\n|\n") == "\n\n\n"
        assert Rle.decode("|||") == "\n\n\n"



import textwrap
from copy import deepcopy

import pytest

from sokoenginepy.game import BoardGraph, Config, HashedBoardManager
from sokoenginepy.io import SokobanPuzzle


@pytest.fixture
def puzzle():
    #   0123456789012345678
    data = """
        ----#####----------
        ----#--@#----------
        ----#$--#----------
        --###--$##---------
        --#--$-$-#---------
        ###-#-##-#---######
        #---#-##-#####--..#
        #-$--$----------..#
        #####-###-#@##--..#
        ----#-----#########
        ----#######--------
    """
    data = textwrap.dedent(data.lstrip("\n").rstrip())
    return SokobanPuzzle(board=data)


@pytest.fixture
def board_graph(puzzle):
    return BoardGraph(puzzle)


class DescribeHashedBoardManager:
    def it_hashes_board_layout(self, board_graph):
        hashed_board_manager = HashedBoardManager(board_graph)

        assert hashed_board_manager.state_hash is not None

    def test_moving_box_modifies_hashes_consistently(self, board_graph):
        hashed_board_manager = HashedBoardManager(board_graph)

        initial_state_hash = hashed_board_manager.state_hash

        initial_box_position = hashed_board_manager.box_position(
            Config.DEFAULT_PIECE_ID
        )
        new_box_position = initial_box_position + 1

        hashed_board_manager.move_box_from(initial_box_position, new_box_position)
        assert hashed_board_manager.state_hash != initial_state_hash

        hashed_board_manager.move_box_from(new_box_position, initial_box_position)
        assert hashed_board_manager.state_hash == initial_state_hash

        hashed_board_manager.move_box(Config.DEFAULT_PIECE_ID, new_box_position)
        assert hashed_board_manager.state_hash != initial_state_hash

        hashed_board_manager.move_box(Config.DEFAULT_PIECE_ID, initial_box_position)
        assert hashed_board_manager.state_hash == initial_state_hash

    def test_moving_pusher_modifies_hashes_consistently(self, board_graph):
        hashed_board_manager = HashedBoardManager(board_graph)

        initial_state_hash = hashed_board_manager.state_hash

        initial_pusher_position = hashed_board_manager.pusher_position(
            Config.DEFAULT_PIECE_ID
        )
        new_pusher_position = initial_pusher_position - 1

        hashed_board_manager.move_pusher_from(
            initial_pusher_position, new_pusher_position
        )
        hashed_board_manager.move_pusher_from(
            new_pusher_position, initial_pusher_position
        )
        assert hashed_board_manager.state_hash == initial_state_hash

        hashed_board_manager.move_pusher(Config.DEFAULT_PIECE_ID, new_pusher_position)
        hashed_board_manager.move_pusher(
            Config.DEFAULT_PIECE_ID, initial_pusher_position
        )
        assert hashed_board_manager.state_hash == initial_state_hash

    def test_setting_boxorder_or_goalorder_on_enabled_sokoban_plus_rehashes_board(
        self, board_graph
    ):
        hashed_board_manager = HashedBoardManager(board_graph)

        initial_state_hash = hashed_board_manager.state_hash
        hashed_board_manager.enable_sokoban_plus()
        assert hashed_board_manager.is_sokoban_plus_enabled is True
        hashed_board_manager.boxorder = "1 2 3"
        assert hashed_board_manager.state_hash != initial_state_hash

        hashed_board_manager.boxorder = None
        initial_state_hash = hashed_board_manager.state_hash
        hashed_board_manager.enable_sokoban_plus()
        assert hashed_board_manager.is_sokoban_plus_enabled is True
        hashed_board_manager.goalorder = "1 2 3"
        assert hashed_board_manager.state_hash != initial_state_hash

    def test_setting_equal_boxorder_or_goalorder_on_enabled_sokoban_plus_doesnt_rehash_board(
        self, board_graph
    ):
        hashed_board_manager = HashedBoardManager(board_graph)

        hashed_board_manager.boxorder = "1 2 3"
        hashed_board_manager.goalorder = "3 2 1"
        hashed_board_manager.enable_sokoban_plus()
        initial_state_hash = hashed_board_manager.state_hash

        hashed_board_manager.boxorder = "1 2 3"
        assert hashed_board_manager.state_hash == initial_state_hash

        hashed_board_manager.goalorder = "3 2 1"
        assert hashed_board_manager.state_hash == initial_state_hash

    def test_setting_boxorder_or_goalorder_on_disabled_sokoban_plus_doesnt_rehash_board(
        self, board_graph
    ):
        hashed_board_manager = HashedBoardManager(board_graph)

        hashed_board_manager.disable_sokoban_plus()
        initial_state_hash = hashed_board_manager.state_hash

        hashed_board_manager.boxorder = "1 2 3"
        assert hashed_board_manager.state_hash == initial_state_hash

        hashed_board_manager.goalorder = "3 2 1"
        assert hashed_board_manager.state_hash == initial_state_hash

    def test_changing_enabled_state_of_sokoban_plus_rehashes_board(self, board_graph):
        hashed_board_manager = HashedBoardManager(board_graph)

        hashed_board_manager.disable_sokoban_plus()

        initial_state_hash = hashed_board_manager.state_hash
        hashed_board_manager.enable_sokoban_plus()
        assert hashed_board_manager.state_hash != initial_state_hash

        initial_state_hash = hashed_board_manager.state_hash
        hashed_board_manager.disable_sokoban_plus()
        assert hashed_board_manager.state_hash != initial_state_hash

    def test_enabling_enabled_sokoban_plus_doesnt_rehash_board(self, board_graph):
        hashed_board_manager = HashedBoardManager(board_graph)

        hashed_board_manager.enable_sokoban_plus()

        initial_state_hash = hashed_board_manager.state_hash
        hashed_board_manager.enable_sokoban_plus()
        assert hashed_board_manager.state_hash == initial_state_hash

    def test_disabling_disabled_sokoban_plus_doesnt_rehash_board(self, board_graph):
        hashed_board_manager = HashedBoardManager(board_graph)

        hashed_board_manager.disable_sokoban_plus()

        initial_state_hash = hashed_board_manager.state_hash
        hashed_board_manager.disable_sokoban_plus()
        assert hashed_board_manager.state_hash == initial_state_hash

    def test_solutions_hashes_are_different_for_enabled_and_disabled_sokoban_plus(
        self, board_graph
    ):
        hashed_board_manager = HashedBoardManager(board_graph)

        disabled_hashes = deepcopy(hashed_board_manager.solutions_hashes)

        hashed_board_manager.boxorder = "1 3 2"
        hashed_board_manager.goalorder = "3 2 1"
        hashed_board_manager.enable_sokoban_plus()

        enabled_hashes = deepcopy(hashed_board_manager.solutions_hashes)

        assert len(disabled_hashes) != 0
        assert len(enabled_hashes) != 0
        assert enabled_hashes != disabled_hashes

    def test_switching_boxes_and_goals_switches_between_positions_hashes(
        self, board_graph
    ):
        hashed_board_manager = HashedBoardManager(board_graph)

        initial_hash = hashed_board_manager.state_hash
        hashed_board_manager.switch_boxes_and_goals()
        after_switch_hash = hashed_board_manager.state_hash
        assert initial_hash != after_switch_hash

        hashed_board_manager.switch_boxes_and_goals()
        assert hashed_board_manager.state_hash == initial_hash
        hashed_board_manager.switch_boxes_and_goals()
        assert hashed_board_manager.state_hash == after_switch_hash
        hashed_board_manager.switch_boxes_and_goals()
        assert hashed_board_manager.state_hash == initial_hash



import pytest

from sokoenginepy.game import Direction, PusherStep
from sokoenginepy.io import SokobanSnapshot


class DescribeSnapshot:
    def it_raises_immediately_on_invalid_characters(self):
        with pytest.raises(ValueError):
            SokobanSnapshot("ZOMG!")

        snapshot = SokobanSnapshot()
        with pytest.raises(ValueError):
            snapshot.moves_data = "ZOMG!"

    def it_raises_delayed_on_invalid_groups(self):
        for data in [
            "lurd[lurd",
            "lurd]lurd",
            "lurd{lurd",
            "lurd}lurd",
        ]:
            snapshot = SokobanSnapshot(data)
            with pytest.raises(ValueError):
                # Trigger delayed parsing
                snapshot.moves_count

    def it_correctly_parses_regular_movement(self):
        data = "lurdLURDL"
        snapshot = SokobanSnapshot(data)

        assert snapshot.moves_count == 4
        assert snapshot.pushes_count == 5
        assert snapshot.jumps_count == 0
        assert not snapshot.is_reverse
        assert str(snapshot) == data
        assert str(snapshot) == snapshot.moves_data

    def it_ignores_spaces_in_input(self):
        data = "lu \t\r\nrdLU   \r\r\n\n\t\tRDL"
        snapshot = SokobanSnapshot(data)

        assert snapshot.moves_count == 4
        assert snapshot.pushes_count == 5
        assert snapshot.jumps_count == 0
        assert not snapshot.is_reverse
        assert str(snapshot) == "lurdLURDL"
        assert snapshot.moves_data == data

    def it_correctly_parses_pusher_selections(self):
        for data in [
            "{lurd}lurdLURDL",
            "lurd{lurd}LURDL",
            "lurdLURDL{lurd}",
        ]:
            snapshot = SokobanSnapshot(data)

            assert snapshot.moves_count == 4
            assert snapshot.pushes_count == 5
            assert snapshot.jumps_count == 0
            assert not snapshot.is_reverse
            assert str(snapshot) == data
            assert str(snapshot) == snapshot.moves_data

        data = "[]lurdLURDD"
        snapshot = SokobanSnapshot(data)
        assert snapshot.moves_count == 4
        assert snapshot.pushes_count == 5
        assert snapshot.jumps_count == 1
        assert snapshot.is_reverse
        assert str(snapshot) == data
        assert str(snapshot) == snapshot.moves_data

    def it_doesnt_allow_pushes_in_pusher_selection(self):
        snapshot = SokobanSnapshot("lurd{LURD}")
        with pytest.raises(ValueError):
            # Trigger delayed parsing
            snapshot.moves_count

    def it_correctly_parses_jumps(self):
        for data in [
            "[lurd]lurdLURDL",
            "lurd[lurd]LURDL",
            "lurdLURDL[lurd]",
        ]:
            snapshot = SokobanSnapshot(data)

            assert snapshot.moves_count == 8
            assert snapshot.pushes_count == 5
            assert snapshot.jumps_count == 1
            assert snapshot.is_reverse

    def it_correctly_prints_jumps(self):
        data = "[lurd]lurdLURDL"
        snapshot = SokobanSnapshot(data)
        assert str(snapshot) == data
        assert str(snapshot) == snapshot.moves_data

        data = "lurd[lurd]LURDL"
        snapshot = SokobanSnapshot(data)
        assert str(snapshot) == "[]" + data
        assert str(snapshot) == "[]" + snapshot.moves_data

        data = "lurdLURDL[lurd]"
        snapshot = SokobanSnapshot(data)
        assert str(snapshot) == "[]" + data
        assert str(snapshot) == "[]" + snapshot.moves_data

    def it_doesnt_allow_pushes_in_jumps(self):
        snapshot = SokobanSnapshot("lurd[LURD]")
        with pytest.raises(ValueError):
            # Trigger delayed parsing
            snapshot.moves_count

    def it_correctly_converts_sequence_of_pusher_steps(self):
        snapshot = SokobanSnapshot()

        steps0 = [
            PusherStep(Direction.LEFT),
            PusherStep(Direction.UP),
            PusherStep(Direction.RIGHT),
            PusherStep(Direction.DOWN),
            PusherStep(Direction.LEFT, box_moved=True),
            PusherStep(Direction.UP, box_moved=True),
            PusherStep(Direction.RIGHT, box_moved=True),
            PusherStep(Direction.DOWN, box_moved=True),
            PusherStep(Direction.DOWN, box_moved=True),
        ]

        steps = steps0
        snapshot.pusher_steps = steps
        assert str(snapshot) == "lurdLURDD"
        assert snapshot.moves_count == 4
        assert snapshot.pushes_count == 5
        assert snapshot.jumps_count == 0
        assert not snapshot.is_reverse

        steps = steps0 + [
            PusherStep(Direction.LEFT, is_jump=True),
            PusherStep(Direction.UP, is_jump=True),
            PusherStep(Direction.RIGHT, is_jump=True),
            PusherStep(Direction.DOWN, is_jump=True),
        ]
        snapshot.pusher_steps = steps
        assert str(snapshot) == "[]lurdLURDD[lurd]"
        assert snapshot.moves_count == 8
        assert snapshot.pushes_count == 5
        assert snapshot.jumps_count == 1
        assert snapshot.is_reverse

        steps = steps0 + [
            PusherStep(Direction.LEFT, is_pusher_selection=True),
            PusherStep(Direction.UP, is_pusher_selection=True),
            PusherStep(Direction.RIGHT, is_pusher_selection=True),
            PusherStep(Direction.DOWN, is_pusher_selection=True),
        ]
        snapshot.pusher_steps = steps
        assert str(snapshot) == "lurdLURDD{lurd}"
        assert snapshot.moves_count == 4
        assert snapshot.pushes_count == 5
        assert snapshot.jumps_count == 0
        assert not snapshot.is_reverse



# import cProfile
from textwrap import dedent

from sokoenginepy import Direction, Mover, SokobanBoard

board = SokobanBoard(
    board_str=dedent(
        """
    ###########
    #       **#
    #       **#
    #  *@   **#
    #       **#
    ###########
"""[
            1:-1
        ]
    )
)

mover = Mover(board)

moves_cycle = 3000 * [
    Direction.LEFT,
    Direction.DOWN,
    Direction.LEFT,
    Direction.LEFT,
    Direction.UP,
    Direction.RIGHT,
    Direction.DOWN,
    Direction.RIGHT,
    Direction.RIGHT,
    Direction.UP,
]


def moves_profile():
    for d in moves_cycle:
        mover.move(d)


def single_move_profile():
    mover.move(Direction.LEFT)


def single_moves_cycle_profile():
    moves_cycle = [
        Direction.LEFT,
        Direction.DOWN,
        Direction.LEFT,
        Direction.LEFT,
        Direction.UP,
        Direction.RIGHT,
        Direction.DOWN,
        Direction.RIGHT,
        Direction.RIGHT,
        Direction.UP,
    ]
    for d in moves_cycle:
        mover.move(d)


if __name__ == "__main__":
    single_moves_cycle_profile()

cProfile.run("main()", "moves_profile.prof")
cProfile.run("mover.move(Direction.LEFT)", "single_move_profile.prof")

# pip install pyprof2calltree
# python bin/mover_profiling.py
# pyprof2calltree -i moves_profile.prof
# pyprof2calltree -i single_move_profile.prof
# kcachegrind moves_profile.prof.log
# kcachegrind single_move_profile.prof.log



from __future__ import annotations

import enum
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, ClassVar, Mapping, Optional, Tuple, Union

from ..io import CellOrientation
from .config import Direction, GraphType
from .pusher_step import PusherStep

if TYPE_CHECKING:
    from .hexoban_tessellation import HexobanTessellation
    from .octoban_tessellation import OctobanTessellation
    from .sokoban_tessellation import SokobanTessellation
    from .trioban_tessellation import TriobanTessellation


class Tessellation(enum.Enum):
    SOKOBAN = 0
    HEXOBAN = 1
    TRIOBAN = 2
    OCTOBAN = 3


def index_1d(x: int, y: int, board_width: int) -> int:
    """Converts 2D coordinate to board position index."""
    return y * board_width + x


def X(index: int, board_width: int) -> int:
    """x component of board position index."""
    return 0 if board_width == 0 else index % board_width


def Y(index: int, board_width: int) -> int:
    """y component of board position index."""
    return 0 if board_width == 0 else int(index / board_width)


def ROW(index: int, board_width: int) -> int:
    """x component of board position index."""
    return Y(index, board_width)


def COLUMN(index: int, board_width: int) -> int:
    """y component of board position index."""
    return X(index, board_width)


def is_on_board_2d(x: int, y: int, board_width: int, board_height: int) -> bool:
    return x >= 0 and y >= 0 and x < board_width and y < board_height


def is_on_board_1d(index: Optional[int], board_width: int, board_height: int) -> bool:
    return (
        index is not None
        and index >= 0
        and is_on_board_2d(
            X(index, board_width), Y(index, board_width), board_width, board_height
        )
    )


class BaseTessellation(metaclass=ABCMeta):
    """
    Base class for all tessellation implementations.
    """

    _LEGAL_DIRECTIONS: ClassVar[Tuple[Direction, ...]] = tuple()
    _CHR_TO_PUSHER_STEP: ClassVar[Mapping[str, Tuple[Direction, bool]]] = {}
    _PUSHER_STEP_TO_CHR: ClassVar[Mapping[Tuple[Direction, bool], str]] = {}

    @classmethod
    def instance(
        cls, tessellation: Tessellation
    ) -> Union[
        TriobanTessellation,
        OctobanTessellation,
        HexobanTessellation,
        SokobanTessellation,
    ]:
        from .hexoban_tessellation import HexobanTessellation
        from .octoban_tessellation import OctobanTessellation
        from .sokoban_tessellation import SokobanTessellation
        from .trioban_tessellation import TriobanTessellation

        if tessellation == Tessellation.SOKOBAN:
            return SokobanTessellation()
        elif tessellation == Tessellation.HEXOBAN:
            return HexobanTessellation()
        elif tessellation == Tessellation.TRIOBAN:
            return TriobanTessellation()
        elif tessellation == Tessellation.OCTOBAN:
            return OctobanTessellation()
        else:
            raise ValueError("Unknown tessellation!")

    @property
    def legal_directions(self) -> Tuple[Direction, ...]:
        """
        Directions that are valid in context of this tessellation.
        """
        return self._LEGAL_DIRECTIONS

    @abstractmethod
    def neighbor_position(
        self, position: int, direction: Direction, board_width: int, board_height: int
    ) -> Optional[int]:
        """
        Calculates neighbor position in given direction.

        Position is always expressed as 1D index of board graph vertex.

        To convert 2D coordinates into vertex index, use :func:`.index_1d` method.

        To convert 1D vertex index into 2D coordinates, use combinations of :func:`.ROW`
        and :func:`.COLUMN` functions.

        Returns:
            int: New position or `None` when new position would be off-board.

        Raises:
            :exc:`ValueError`: ``direction`` is not one of :attr:`legal_directions`
        """
        pass

    @property
    def graph_type(self) -> GraphType:
        """
        Type of board graph used in context of this tessellation.
        """
        return GraphType.DIRECTED

    def pusher_step_to_char(self, pusher_step: PusherStep) -> str:
        """
        Converts :class:`.PusherStep` to movement character.

        Raises:
            :exc:`ValueError`: conversion not possible in context of this tessellation
        """
        try:
            retv = self._PUSHER_STEP_TO_CHR[
                (pusher_step.direction, pusher_step.is_push_or_pull)
            ]
        except KeyError:
            raise ValueError(pusher_step)

        return retv

    def char_to_pusher_step(self, input_chr: str) -> PusherStep:
        """
        Converts movement character to :class:`.PusherStep`.

        Raises:
            :exc:`ValueError`: conversion is not possible in context of this
                               tessellation
        """
        try:
            direction, box_moved = self._CHR_TO_PUSHER_STEP[input_chr]
        except KeyError:
            raise ValueError(input_chr)

        return PusherStep(direction=direction, box_moved=box_moved)

    def cell_orientation(
        self, position: int, board_width: int, board_height: int
    ) -> CellOrientation:
        """
        Calculates board cell orientation for given coordinate.
        """
        return CellOrientation.DEFAULT



from copy import deepcopy
from itertools import permutations

import pytest

from sokoenginepy.game import (
    Config,
    PusherStep,
    BoardGraph,
    Direction,
    IllegalMoveError,
    Mover,
    NonPlayableBoardError,
    SolvingMode,
    index_1d,
)
from sokoenginepy.io import SokobanPuzzle


@pytest.fixture
def non_playable_board():
    return BoardGraph(SokobanPuzzle(5, 5))


@pytest.fixture
def off_board_position():
    return index_1d(42, 42, 42)


@pytest.fixture
def forward_board():
    data = "\n".join(
        [
            # 12345678
            "#########",  # 0
            "#$  .  .#",  # 1
            "#   @$# #",  # 2
            "#.$    @#",  # 3
            "#########",  # 4
        ]
    )
    return BoardGraph(SokobanPuzzle(board=data))


@pytest.fixture
def reverse_board():
    data = "\n".join(
        [
            # 12345678
            "#########",  # 0
            "#.  $  $#",  # 1
            "#   @.# #",  # 2
            "#$.    @#",  # 3
            "#########",  # 4
        ]
    )
    return BoardGraph(SokobanPuzzle(board=data))


@pytest.fixture
def forward_mover_moves_cycle():
    return [
        Direction.LEFT,
        Direction.LEFT,
        Direction.LEFT,
        Direction.DOWN,
        Direction.RIGHT,
        Direction.UP,
        Direction.RIGHT,
        Direction.RIGHT,
        Direction.DOWN,
        Direction.LEFT,
        Direction.UP,
        Direction.RIGHT,
    ]


@pytest.fixture
def reverse_mover_moves_cycle():
    return [
        Direction.LEFT,
        Direction.UP,
        Direction.LEFT,
        Direction.DOWN,
        Direction.RIGHT,
        Direction.RIGHT,
        Direction.UP,
        Direction.RIGHT,
        Direction.DOWN,
        Direction.LEFT,
    ]


@pytest.fixture
def jumps():
    return [
        [PusherStep(direction, is_jump=True) for direction in permutation]
        for permutation in permutations(
            [Direction.UP, Direction.LEFT, Direction.LEFT, Direction.LEFT]
        )
    ]


@pytest.fixture
def undone_jumps():
    return [
        [PusherStep(direction, is_jump=True) for direction in permutation]
        for permutation in permutations(
            [Direction.DOWN, Direction.RIGHT, Direction.RIGHT, Direction.RIGHT]
        )
    ]


@pytest.fixture
def pusher_selections():
    return [
        [PusherStep(direction, is_pusher_selection=True) for direction in permutation]
        for permutation in permutations(
            [Direction.DOWN, Direction.RIGHT, Direction.RIGHT, Direction.RIGHT]
        )
    ]


@pytest.fixture
def undone_pusher_selections():
    return [
        [PusherStep(direction, is_pusher_selection=True) for direction in permutation]
        for permutation in permutations(
            [Direction.UP, Direction.LEFT, Direction.LEFT, Direction.LEFT]
        )
    ]


@pytest.fixture
def jump_dest(reverse_board):
    return index_1d(1, 1, reverse_board.board_width)


@pytest.fixture
def jump_obstacle_position(reverse_board):
    return index_1d(0, 0, reverse_board.board_width)


class DescribeMover:
    def it_raises_if_board_is_not_playable(self, non_playable_board):
        with pytest.raises(NonPlayableBoardError):
            mover = Mover(non_playable_board)

    def it_assumes_forward_solving_mode(self, forward_board):
        mover = Mover(forward_board)
        assert mover.solving_mode == SolvingMode.FORWARD

    def it_switches_boxes_and_goals_if_reverse_solving_mode_requested(
        self, forward_board, reverse_board
    ):
        mover = Mover(forward_board, SolvingMode.REVERSE)
        assert str(mover.board) == str(reverse_board)

    class DescribeSelectingPusher:
        def it_pre_selects_first_pusher(self, forward_board):
            mover = Mover(forward_board)
            assert mover.selected_pusher == Config.DEFAULT_PIECE_ID

        def it_can_select_pusher_that_will_perform_next_move(self, forward_board):
            forward_mover = Mover(forward_board)

            forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID + 1
            for am in forward_mover.last_move:
                assert am.pusher_id == Config.DEFAULT_PIECE_ID
                assert am.moved_box_id == None

        def it_raises_if_trying_to_select_non_existent_pusher(self, forward_board):
            forward_mover = Mover(forward_board)

            with pytest.raises(KeyError):
                forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 42)

        def it_updates_last_move_with_pusher_selection_sequence(
            self, forward_board, pusher_selections
        ):
            forward_mover = Mover(forward_board)

            forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            assert forward_mover.last_move in pusher_selections

        def when_re_selecting_same_pusher_it_doesnt_update_last_move(
            self, forward_board, pusher_selections
        ):
            forward_mover = Mover(forward_board)

            forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            last_move = deepcopy(forward_mover.last_move)
            assert last_move in pusher_selections
            forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            assert forward_mover.last_move == last_move

        def it_doesnt_update_last_move_for_failed_pusher_selection(self, forward_board):
            forward_mover = Mover(forward_board)

            last_move = deepcopy(forward_mover.last_move)
            with pytest.raises(KeyError):
                forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 2)
            assert forward_mover.last_move == last_move

            forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            last_move = deepcopy(forward_mover.last_move)
            with pytest.raises(KeyError):
                forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 2)
            assert forward_mover.last_move == last_move

        def it_undoes_pusher_selection_and_updates_last_move(
            self, forward_board, undone_pusher_selections
        ):
            forward_mover = Mover(forward_board)

            assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID
            forward_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID + 1
            forward_mover.undo_last_move()

            assert forward_mover.selected_pusher == Config.DEFAULT_PIECE_ID
            assert forward_mover.last_move in undone_pusher_selections
            for am in forward_mover.last_move:
                assert am.pusher_id == Config.DEFAULT_PIECE_ID
                assert am.moved_box_id == None

    class DescribeJumping:
        def it_performs_jumps(self, forward_board, jump_dest):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            reverse_mover.jump(jump_dest)
            assert (
                reverse_mover.board_manager.pusher_position(Config.DEFAULT_PIECE_ID)
                == jump_dest
            )
            for am in reverse_mover.last_move:
                assert am.pusher_id == Config.DEFAULT_PIECE_ID
                assert am.moved_box_id == None

        def it_refuses_to_jump_in_forward_solving_mode(self, forward_board, jump_dest):
            forward_mover = Mover(forward_board)

            with pytest.raises(IllegalMoveError):
                forward_mover.jump(jump_dest)
            assert not forward_mover.last_move

        def it_refuses_to_jump_after_first_pull(self, forward_board, jump_dest):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            reverse_mover.pulls_boxes = True
            reverse_mover.move(Direction.DOWN)
            reverse_mover.last_move = None
            with pytest.raises(IllegalMoveError):
                reverse_mover.jump(jump_dest)
            assert not reverse_mover.last_move

        def it_allows_jumps_after_first_pull_is_undone(self, forward_board, jump_dest):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            reverse_mover.pulls_boxes = True
            reverse_mover.move(Direction.DOWN)
            assert reverse_mover.last_move[0].is_push_or_pull
            reverse_mover.undo_last_move()
            reverse_mover.jump(jump_dest)
            assert reverse_mover.board[jump_dest].has_pusher
            for am in reverse_mover.last_move:
                assert am.pusher_id == Config.DEFAULT_PIECE_ID
                assert am.moved_box_id == None

        def it_refuses_to_jump_onto_obstacles(
            self, forward_board, jump_obstacle_position
        ):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            with pytest.raises(IllegalMoveError):
                reverse_mover.jump(jump_obstacle_position)
            assert not reverse_mover.last_move

        def it_refuses_to_jump_off_the_board(self, forward_board, off_board_position):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            with pytest.raises(IndexError):
                reverse_mover.jump(off_board_position)
            assert not reverse_mover.last_move

        def it_updates_last_move_with_jump_sequence(
            self, forward_board, jump_dest, jumps
        ):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            reverse_mover.jump(jump_dest)
            assert reverse_mover.last_move in jumps
            for am in reverse_mover.last_move:
                assert am.pusher_id == Config.DEFAULT_PIECE_ID
                assert am.moved_box_id == None

        def when_jumping_to_same_position_it_doesnt_update_last_move(
            self, forward_board, jump_dest
        ):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            reverse_mover.jump(jump_dest)
            reverse_mover.last_move = None
            reverse_mover.jump(jump_dest)
            assert not reverse_mover.last_move

        def it_doesnt_update_last_move_for_failed_jumps(
            self, forward_board, jump_dest, jump_obstacle_position
        ):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            reverse_mover.jump(jump_dest)
            last_move = deepcopy(reverse_mover.last_move)
            with pytest.raises(IllegalMoveError):
                reverse_mover.jump(jump_obstacle_position)
            assert reverse_mover.last_move == last_move

        def it_raises_when_undoing_jump_for_forward_mover(self, forward_board, jumps):
            forward_mover = Mover(forward_board)

            forward_mover.last_move = jumps[0]
            with pytest.raises(IllegalMoveError):
                forward_mover.undo_last_move()
            assert forward_mover.last_move == jumps[0]

        def it_undoes_jump_and_updates_last_move(
            self, forward_board, jump_dest, undone_jumps
        ):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            src = index_1d(4, 2, reverse_mover.board.board_width)
            reverse_mover.jump(jump_dest)

            reverse_mover.undo_last_move()
            assert reverse_mover.board_manager.pusher_position(Config.DEFAULT_PIECE_ID) == src
            assert reverse_mover.board[src].has_pusher is True
            assert reverse_mover.board[jump_dest].has_pusher is False

            assert reverse_mover.last_move in undone_jumps

    class DescribeForwardMovement:
        def it_forward_moves_pusher_in_requested_direction(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @ $. #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(4, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board)
            mover.select_pusher(selected_pusher)

            mover.move(Direction.RIGHT)
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.RIGHT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def when_forward_moving_pusher_doesnt_pull_box_behind_it(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "# $@ .  #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(4, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board)
            mover.select_pusher(selected_pusher)

            mover.move(Direction.RIGHT)
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.RIGHT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def it_refuses_to_forward_move_pusher_into_obstacles_or_off_board(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "@$   .  #",  # 2
                            "@       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(0, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board)
            mover.select_pusher(selected_pusher)
            last_move = mover.last_move

            for direction in [Direction.UP, Direction.DOWN, Direction.LEFT]:
                with pytest.raises(IllegalMoveError):
                    mover.move(direction)
                assert mover.board_manager.pusher_position(selected_pusher) == src
                assert board[src].has_pusher
                assert mover.last_move == last_move

        def it_undoes_forward_move(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @ $. #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(2, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.RIGHT)]
            mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.LEFT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def when_undoing_forward_move_doesnt_pull_box_from_behind_pusher(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @$.  #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(2, 2, board.board_width)
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.RIGHT)]
            mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[src + 1].has_box
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.LEFT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def it_refuses_to_undo_forward_move_by_moving_pusher_into_obstacles_or_off_board(
            self,
        ):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "@$   .  #",  # 2
                            "@       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(0, 2, board.board_width)
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            for direction in [
                Direction.DOWN,
                Direction.UP,
                Direction.LEFT,
                Direction.RIGHT,
            ]:
                mover.last_move = [PusherStep(direction)]
                with pytest.raises(IllegalMoveError):
                    mover.undo_last_move()
                assert mover.board_manager.pusher_position(selected_pusher) == src
                assert board[src].has_pusher
                # last move shuould be unchanged
                assert mover.last_move == [PusherStep(direction)]
                assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
                assert mover.last_move[0].moved_box_id == None

        def it_pushes_box_in_front_of_pusher(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#   @$. #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            box_src = index_1d(5, 2, board.board_width)
            box_dest = index_1d(6, 2, board.board_width)
            pusher_src = index_1d(4, 2, board.board_width)
            pusher_dest = box_src
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            mover.move(Direction.RIGHT)

            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID) == box_dest
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert not mover.board[box_src].has_box
            assert not mover.board[pusher_src].has_pusher
            assert mover.board[box_dest].has_box
            assert mover.board[pusher_dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID

        def when_pushing_box_doesnt_pull_box_from_behind_pusher(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  $@$..#",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            box_src = index_1d(5, 2, board.board_width)
            box_dest = index_1d(6, 2, board.board_width)
            pusher_src = index_1d(4, 2, board.board_width)
            pusher_dest = box_src
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            mover.move(Direction.RIGHT)

            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID + 1) == box_dest
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert not mover.board[box_src].has_box
            assert not mover.board[pusher_src].has_pusher
            assert mover.board[box_dest].has_box
            assert mover.board[pusher_dest].has_pusher
            assert not mover.board[pusher_src].has_box
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID + 1

        def it_refuses_to_push_box_into_obstacles_or_off_board(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#####$###",  # 0
                            "#@ $$@$##",  # 1
                            "#..  $  #",  # 2
                            "#... @  #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            pusher_src = index_1d(5, 1, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board)
            mover.select_pusher(selected_pusher)
            last_move = mover.last_move

            for direction in [
                Direction.UP,
                Direction.DOWN,
                Direction.LEFT,
                Direction.RIGHT,
            ]:
                with pytest.raises(IllegalMoveError):
                    mover.move(direction)

                assert (
                    mover.board_manager.pusher_position(selected_pusher) == pusher_src
                )
                assert board[pusher_src].has_pusher
                assert board[board.neighbor(pusher_src, direction)].has_box
                assert board[board.neighbor(pusher_src, direction.opposite)].has_box
                assert mover.last_move == last_move

        def it_undoes_push(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#   @$. #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            box_src = index_1d(5, 2, board.board_width)
            pusher_src = index_1d(4, 2, board.board_width)
            pusher_dest = pusher_src - 1
            box_dest = box_src - 1
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.RIGHT, box_moved=True)]
            mover.undo_last_move()

            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID) == box_dest
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert not mover.board[box_src].has_box
            assert not mover.board[pusher_src].has_pusher
            assert mover.board[box_dest].has_box
            assert mover.board[pusher_dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.LEFT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID

        def it_refuses_to_undo_push_moving_pusher_into_obstacles_or_off_board(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@   $..#",  # 1
                            "@$  @@$.#",  # 2
                            "$    $..#",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            pusher_src = index_1d(0, 2, board.board_width)
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            # Undo into off board
            mover.last_move = [PusherStep(Direction.RIGHT, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_src
            assert board[pusher_src].has_pusher
            assert board[board.neighbor(pusher_src, Direction.RIGHT)].has_box
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

            # Undo into wall
            mover.last_move = [PusherStep(Direction.DOWN, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_src
            assert board[pusher_src].has_pusher
            assert board[board.neighbor(pusher_src, Direction.DOWN)].has_box
            assert mover.last_move == [PusherStep(Direction.DOWN, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

            selected_pusher = Config.DEFAULT_PIECE_ID + 3
            mover.select_pusher(selected_pusher)
            pusher_src = index_1d(5, 2, board.board_width)

            # Undo into pusher
            mover.last_move = [PusherStep(Direction.RIGHT, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_src
            assert board[pusher_src].has_pusher
            assert board[board.neighbor(pusher_src, Direction.RIGHT)].has_box
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

            # Undo into box
            mover.last_move = [PusherStep(Direction.DOWN, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == pusher_src
            assert board[pusher_src].has_pusher
            assert board[board.neighbor(pusher_src, Direction.DOWN)].has_box
            assert mover.last_move == [PusherStep(Direction.DOWN, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

        def it_refuses_to_undo_push_if_there_is_no_box_behind_pusher(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#   @ .$#",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            mover = Mover(board)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.RIGHT, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]

    class DescribeReverseMovement:
        def it_moves_pusher_in_requested_direction(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @ $. #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(4, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.move(Direction.RIGHT)
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.RIGHT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def when_reverse_moving_it_doesnt_pull_box_behind_pusher(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @. $ #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(2, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)
            mover.pulls_boxes = False

            mover.move(Direction.LEFT)
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.LEFT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def it_refuses_to_reverse_move_pusher_into_obstacles_or_off_board(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "@.   $  #",  # 2
                            "@       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(0, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)
            last_move = mover.last_move

            for direction in [
                Direction.UP,
                Direction.DOWN,
                Direction.LEFT,
                Direction.RIGHT,
            ]:
                with pytest.raises(IllegalMoveError):
                    mover.move(direction)
                assert mover.board_manager.pusher_position(selected_pusher) == src
                assert board[src].has_pusher
                assert mover.last_move == last_move

        def it_undoes_reverse_move(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @ $. #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(4, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.LEFT)]
            mover.undo_last_move()
            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.RIGHT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def when_undoing_reverse_move_it_doesnt_pull_box_behind_pusher(self):
            board_str = "\n".join(
                [
                    # 12345678
                    "#########",  # 0
                    "#@      #",  # 1
                    "#  @.$  #",  # 2
                    "#       #",  # 3
                    "#########",  # 4
                ]
            )

            # pulls_boxes == False
            board = BoardGraph(SokobanPuzzle(board=board_str))
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(2, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.RIGHT)]
            mover.pulls_boxes = False
            mover.undo_last_move()

            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert not board[src].has_box
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.LEFT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

            # pulls_boxes == True
            board = BoardGraph(SokobanPuzzle(board=board_str))
            src = index_1d(3, 2, board.board_width)
            dest = index_1d(2, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.RIGHT)]
            mover.pulls_boxes = True
            mover.undo_last_move()

            assert mover.board_manager.pusher_position(selected_pusher) == dest
            assert not board[src].has_pusher
            assert not board[src].has_box
            assert board[dest].has_pusher
            assert mover.last_move == [PusherStep(Direction.LEFT)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == None

        def it_pulls_box_behind_pusher_if_pulls_boxes_is_set(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @.$  #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            pusher_src = index_1d(3, 2, board.board_width)
            pusher_dest = index_1d(2, 2, board.board_width)
            box_src = index_1d(4, 2, board.board_width)
            box_dest = pusher_src
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.pulls_boxes = True
            mover.move(Direction.LEFT)

            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID) == box_dest
            assert board[pusher_dest].has_pusher
            assert not board[pusher_src].has_pusher
            assert board[box_dest].has_box
            assert not board[box_src].has_box
            assert mover.last_move == [PusherStep(Direction.LEFT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID

        def it_refuses_to_pull_box_if_pusher_would_move_into_obstacle_or_off_board(
            self,
        ):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@ $  .$#",  # 1
                            "@. $ @@.#",  # 2
                            ".  $  .$#",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            pusher_src = index_1d(3, 2, board.board_width)
            mover = Mover(board, SolvingMode.REVERSE)
            mover.pulls_boxes = True
            mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            mover.last_move = []

            with pytest.raises(IllegalMoveError):
                mover.move(Direction.LEFT)
            assert mover.last_move == []

            with pytest.raises(IllegalMoveError):
                mover.move(Direction.UP)
            assert mover.last_move == []

            mover.select_pusher(Config.DEFAULT_PIECE_ID + 3)
            mover.last_move = []

            with pytest.raises(IllegalMoveError):
                mover.move(Direction.LEFT)
            assert mover.last_move == []

            with pytest.raises(IllegalMoveError):
                mover.move(Direction.UP)
            assert mover.last_move == []

        def it_undoes_pull(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@      #",  # 1
                            "#  @.$  #",  # 2
                            "#       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            pusher_src = index_1d(3, 2, board.board_width)
            pusher_dest = index_1d(4, 2, board.board_width)
            box_src = pusher_dest
            box_dest = index_1d(5, 2, board.board_width)
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.LEFT, box_moved=True)]
            mover.undo_last_move()

            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID) == box_dest
            assert board[pusher_dest].has_pusher
            assert not board[pusher_src].has_pusher
            assert board[box_dest].has_box
            assert not board[box_src].has_box
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID

        def it_undoes_pull_not_moving_box_behind_pusher(self):
            board_str = "\n".join(
                [
                    # 12345678
                    "#########",  # 0
                    "#@     $#",  # 1
                    "# .@.$  #",  # 2
                    "#       #",  # 3
                    "#########",  # 4
                ]
            )

            # pulls_boxes == False
            board = BoardGraph(SokobanPuzzle(board=board_str))
            pusher_src = index_1d(3, 2, board.board_width)
            pusher_dest = index_1d(4, 2, board.board_width)
            box_src = pusher_dest
            box_dest = index_1d(5, 2, board.board_width)
            behind_pusher = pusher_src - 1
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.LEFT, box_moved=True)]
            mover.pulls_boxes = False
            mover.undo_last_move()

            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID + 1) == box_dest
            assert board[pusher_dest].has_pusher
            assert not board[pusher_src].has_pusher
            assert board[box_dest].has_box
            assert not board[box_src].has_box
            assert board[behind_pusher].has_box
            assert not board[pusher_src].has_box
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID + 1

            # pulls_boxes == True
            board = BoardGraph(SokobanPuzzle(board=board_str))
            pusher_src = index_1d(3, 2, board.board_width)
            pusher_dest = index_1d(4, 2, board.board_width)
            box_src = pusher_dest
            box_dest = index_1d(5, 2, board.board_width)
            behind_pusher = pusher_src - 1
            selected_pusher = Config.DEFAULT_PIECE_ID + 1
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(selected_pusher)

            mover.last_move = [PusherStep(Direction.LEFT, box_moved=True)]
            mover.pulls_boxes = True
            mover.undo_last_move()

            assert mover.board_manager.pusher_position(selected_pusher) == pusher_dest
            assert mover.board_manager.box_position(Config.DEFAULT_PIECE_ID + 1) == box_dest
            assert board[pusher_dest].has_pusher
            assert not board[pusher_src].has_pusher
            assert board[box_dest].has_box
            assert not board[box_src].has_box
            assert board[behind_pusher].has_box
            assert not board[pusher_src].has_box
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == selected_pusher
            assert mover.last_move[0].moved_box_id == Config.DEFAULT_PIECE_ID + 1

        def it_refuses_to_undo_pull_moving_into_obstacles_or_off_board(self):
            board = BoardGraph(
                SokobanPuzzle(
                    board="\n".join(
                        [
                            # 12345678
                            "#########",  # 0
                            "#@ $$$$$#",  # 1
                            "@. ..@.@#",  # 2
                            ".       #",  # 3
                            "#########",  # 4
                        ]
                    )
                )
            )
            mover = Mover(board, SolvingMode.REVERSE)
            mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)

            # undo into off board
            mover.last_move = [PusherStep(Direction.RIGHT, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

            # undo into wall
            mover.last_move = [PusherStep(Direction.DOWN, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.last_move == [PusherStep(Direction.DOWN, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

            mover.select_pusher(Config.DEFAULT_PIECE_ID + 2)

            # undo into box
            mover.last_move = [PusherStep(Direction.RIGHT, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.last_move == [PusherStep(Direction.RIGHT, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

            # undo into pusher
            mover.last_move = [PusherStep(Direction.LEFT, box_moved=True)]
            with pytest.raises(IllegalMoveError):
                mover.undo_last_move()
            assert mover.last_move == [PusherStep(Direction.LEFT, box_moved=True)]
            assert mover.last_move[0].pusher_id == Config.DEFAULT_PIECE_ID
            assert mover.last_move[0].moved_box_id == None

    class DescribeUndoLastMove:
        def it_can_undo_random_sequence_of_moves_stored_in_last_move(
            self, forward_board, jump_dest, undone_jumps
        ):
            reverse_mover = Mover(forward_board, SolvingMode.REVERSE)

            initial_board = str(reverse_mover.board)
            moves = []
            reverse_mover.jump(jump_dest)
            moves += reverse_mover.last_move
            jump = deepcopy(reverse_mover.last_move)

            reverse_mover.select_pusher(Config.DEFAULT_PIECE_ID + 1)
            moves += reverse_mover.last_move
            selection = deepcopy(reverse_mover.last_move)
            undone_selections = [
                [
                    PusherStep(
                        direction=am.direction.opposite, is_pusher_selection=True
                    )
                    for am in permutation
                ]
                for permutation in permutations(selection)
            ]

            reverse_mover.move(Direction.UP)
            moves += reverse_mover.last_move
            reverse_mover.move(Direction.DOWN)
            moves += reverse_mover.last_move
            reverse_mover.move(Direction.LEFT)
            moves += reverse_mover.last_move

            board_after_moving = str(reverse_mover.board)

            reverse_mover.last_move = moves

            reverse_mover.undo_last_move()

            assert reverse_mover.last_move[0] == PusherStep(Direction.RIGHT)
            assert reverse_mover.last_move[1] == PusherStep(
                Direction.UP, box_moved=True
            )
            assert reverse_mover.last_move[2] == PusherStep(Direction.DOWN)
            assert reverse_mover.last_move[3 : 3 + len(selection)] in undone_selections
            assert reverse_mover.last_move[3 + len(selection) :] in undone_jumps

            assert str(reverse_mover.board) == initial_board

        def when_undoing_moves_it_raises_on_first_illegal_move(self):
            # TODO
            pass

        def when_fails_it_leaves_successful_moves_in_last_move(self):
            # TODO
            pass



import json
import os

import pytest

from sokoenginepy.io import HexobanPuzzle


@pytest.fixture(scope="session")
def tests_data(resources_root):
    path = os.path.join(resources_root, "test_data", "hexoban_parser_tests.json")

    with open(path) as f:
        data = json.load(f)

    return data


class DescribeHexobanPuzzle:
    def it_raises_on_illegal_scheme(self, tests_data):
        board = "\n".join(tests_data["illegal_scheme1"])
        puzzle = HexobanPuzzle(board=board)
        with pytest.raises(ValueError):
            # Trigger parsing
            puzzle.width

        board = "\n".join(tests_data["illegal_scheme2"])
        puzzle = HexobanPuzzle(board=board)
        with pytest.raises(ValueError):
            # Trigger parsing
            puzzle.width

    class describe_parser_tests:
        def it_correctly_parses_hexocet_collection(self, tests_data):
            for level_title, test_case in tests_data[_PARSING_HEXOCET_KEY].items():
                src = "\n".join(test_case["src"])
                parsed = "\n".join(test_case["parsed"])
                parsed_width = test_case["parsed_width"]
                parsed_height = test_case["parsed_height"]
                fail_msg = f"Failed Hexocet parsing test: {level_title}"

                raised = False
                try:
                    in_puzzle = HexobanPuzzle(board=src)
                    in_puzzle.width
                    out_puzzle = HexobanPuzzle(board=parsed)
                    out_puzzle.width
                except ValueError:
                    raised = True

                assert not raised, fail_msg
                assert in_puzzle.board == src, fail_msg
                assert out_puzzle.board == parsed, fail_msg
                assert str(in_puzzle) == parsed, fail_msg
                assert str(in_puzzle) == str(out_puzzle), fail_msg
                assert in_puzzle.width == parsed_width, fail_msg
                assert in_puzzle.height == parsed_height, fail_msg

        def it_correctly_parses_various_hexoban_text_layout_schemes(self, tests_data):
            for level_title, test_case in tests_data[_PARSING_SCHEMES_KEY].items():
                src = "\n".join(test_case["src"])
                parsed = "\n".join(test_case["parsed"])
                parsed_width = test_case["parsed_width"]
                parsed_height = test_case["parsed_height"]
                fail_msg = f"Failed schemes parsing test: {level_title}"

                raised = False
                try:
                    in_puzzle = HexobanPuzzle(board=src)
                    in_puzzle.width
                    out_puzzle = HexobanPuzzle(board=parsed)
                    out_puzzle.width
                except ValueError:
                    raised = True

                assert not raised, fail_msg
                assert in_puzzle.board == src, fail_msg
                assert out_puzzle.board == parsed, fail_msg
                assert str(in_puzzle) == parsed, fail_msg
                assert str(in_puzzle) == str(out_puzzle), fail_msg
                assert in_puzzle.width == parsed_width, fail_msg
                assert in_puzzle.height == parsed_height, fail_msg

    class describe_row_and_column_reordering:
        def it_correctly_reorders_rows_and_columns(self, tests_data):
            for test_title, test_case in tests_data[_REVERSING_KEY].items():
                src_key = test_case["src"]
                board = "\n".join(tests_data[_PARSING_SCHEMES_KEY][src_key]["src"])
                result = "\n".join(test_case["result"])
                method_name = test_case["method"]
                fail_msg = f'Failed mirroring test "{test_title}"!'

                raised = False
                try:
                    puzzle = HexobanPuzzle(board=board)
                    getattr(puzzle, method_name)()
                except ValueError:
                    raised = True

                assert not raised, fail_msg
                assert str(puzzle) == result, fail_msg
                assert puzzle.width == test_case["result_width"], fail_msg
                assert puzzle.height == test_case["result_height"], fail_msg

    class describe_board_resizing:
        def it_can_resize_board_independently_of_encoding_scheme(self, tests_data):
            for test_title, test_case in tests_data[_RESIZING_KEY].items():
                src_key = test_case["src"]
                board = "\n".join(tests_data[_PARSING_SCHEMES_KEY][src_key]["src"])
                result = "\n".join(test_case["result"])
                method_name = test_case["method"]
                fail_msg = f'Failed resizing test "{test_title}"!'

                raised = False
                try:
                    puzzle = HexobanPuzzle(board=board)
                    puzzle.width
                    getattr(puzzle, method_name)()
                except ValueError:
                    raised = True

                assert not raised, fail_msg
                assert str(puzzle) == result, fail_msg
                assert puzzle.width == test_case["result_width"], fail_msg
                assert puzzle.height == test_case["result_height"], fail_msg

        def test_resizes_board_to_bigger(self, tests_data):
            board = "\n".join(tests_data[_PARSING_SCHEMES_KEY]["scheme1_type1"]["src"])
            puzzle = HexobanPuzzle(board=board)
            old_width = puzzle.width
            old_height = puzzle.height
            puzzle.resize(puzzle.width + 5, puzzle.height + 5)
            assert puzzle.width == old_width + 5
            assert puzzle.height == old_height + 5

        def test_resizes_board_to_smaller(self, tests_data):
            board = "\n".join(tests_data[_PARSING_SCHEMES_KEY]["scheme1_type1"]["src"])
            puzzle = HexobanPuzzle(board=board)
            old_height = puzzle.height
            puzzle.resize(puzzle.width - 3, puzzle.height - 3)
            assert puzzle.width == 7
            assert puzzle.height == old_height - 3


_PARSING_HEXOCET_KEY = "parsing - Hexocet"
_PARSING_SCHEMES_KEY = "parsing - schemes"
_RESIZING_KEY = "resizing"
_REVERSING_KEY = "reversing"
